# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
from moleculekit.projections.projection import Projection
import numpy as np
import logging

logger = logging.getLogger(__name__)


class AtomNotFoundException(Exception):
    pass


class Dihedral:
    """Class to store atoms defining a dihedral angle.

    Example
    -------
    >>> # Using the helper functions to construct a dihedral object
    >>> d1 = Dihedral.phi(mol, 5, 6, segid='P0')
    >>> d2 = Dihedral.chi1(mol, 12, segid='P0') # Defining segid
    >>> d3 = Dihedral.chi1(mol, 38, chain='X') # Defining chain
    >>> # Manual construction
    >>> atom1 = {'name': 'N', 'resid': 5, 'segid': 'P'}
    >>> atom2 = {'name': 'CA', 'resid': 3, 'segid': 'P', 'chain': 'A', 'insertion': 'B'}
    >>> atom3 = {'name': 'C', 'resid': 46, 'segid': 'P', 'chain': 'X'}
    >>> atom4 = {'name': 'O', 'resid': 2, 'segid': 'P'}
    >>> d = Dihedral(atom1, atom2, atom3, atom4)
    """

    def __init__(self, atom1, atom2, atom3, atom4, dihedraltype=None, check_valid=True):
        if check_valid:
            valid_keys = ("name", "resid", "segid", "insertion", "chain")
            default_values = {
                "name": "",
                "resid": 0,
                "segid": "",
                "insertion": "",
                "chain": "",
            }
            for i, a in enumerate([atom1, atom2, atom3, atom4]):
                for k in a.keys():
                    if k not in valid_keys:
                        raise RuntimeError(
                            'Dictionary key can\'t be "{}". Valid keys are: {}'.format(
                                k, valid_keys
                            )
                        )
                for k in np.setdiff1d(valid_keys, tuple(a.keys())):
                    print(np.setdiff1d(valid_keys, tuple(a.keys())))
                    a[k] = default_values[k]
                # frames.append(pd.DataFrame(a, index=[i]))
        # self.atoms = pd.concat(frames)
        self.atoms = [atom1, atom2, atom3, atom4]
        self.dihedraltype = dihedraltype

    def __str__(self):
        descr = ""
        if self.dihedraltype is not None:
            descr = f'"{self.dihedraltype}" dihedral angle including atoms:\n'
        descr += "name\tresid\tinsertion\tchain\tsegid\n"
        for a in self.atoms:
            descr += "{}\t{}\t{}\t\t{}\t{}\n".format(
                a["name"], a["resid"], a["insertion"], a["chain"], a["segid"]
            )
        return descr

    def __repr__(self):
        return self.__str__()

    def _explanation(self):
        descr = ""
        if self.dihedraltype is not None:
            descr += self.dihedraltype + " "
        resids = []
        insertions = []
        for a in self.atoms:
            resids.append(a["resid"])
            insertions.append(a["insertion"])
        uqresids = list(np.unique(resids))
        uqinsertions = list(np.unique(insertions))

        descr += "dihedral of Resid/Insertion/Chain/Segid/ {}/{}/{}/{}".format(
            ",".join(map(str, uqresids)),
            ",".join(uqinsertions),
            self.atoms[0]["chain"],
            self.atoms[0]["segid"],
        )
        return descr

    @staticmethod
    def dihedralsToIndexes(mol, dihedrals, sel="all"):
        """Converts dihedral objects to atom indexes of a given Molecule

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain atom information
        dihedrals : list
            A single dihedral or a list of Dihedral objects
        sel : str
            Atom selection string to restrict the application of the selections.
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__

        Returns
        -------
        indexes : list of lists
            A list containing a list of atoms that correspond to each dihedral.

        Examples
        --------
        >>> dihs = []
        >>> dihs.append(Dihedral.phi(mol, 1, 2))
        >>> dihs.append(Dihedral.psi(mol, 2, 3))
        >>> indexes = Dihedral.dihedralsToIndexes(mol, dihs)
        """
        selatoms = mol.atomselect(sel)
        from moleculekit.util import ensurelist

        indexes = []
        for dih in ensurelist(dihedrals):
            idx = []
            for a in dih.atoms:
                atomsel = (
                    (mol.name == a["name"])
                    & (mol.resid == a["resid"])
                    & (mol.insertion == a["insertion"])
                    & (mol.chain == a["chain"])
                    & (mol.segid == a["segid"])
                )
                atomsel = atomsel & selatoms
                if np.sum(atomsel) != 1:
                    raise RuntimeError(
                        "Expected one atom from atomselection {}. Got {} instead.".format(
                            a, np.sum(atomsel)
                        )
                    )
                idx.append(np.where(atomsel)[0][0])
            indexes.append(idx)
        return indexes

    @staticmethod
    def _findResidue(mol, resid, insertion=None, chain=None, segid=None):
        idx = mol.resid == resid
        descr = f'Resid "{resid}"'
        if insertion is not None:
            idx &= mol.insertion == insertion
            descr += f' Insertion "{insertion}"'
        if chain is not None:
            idx &= mol.chain == chain
            descr += f' Chain "{chain}"'
        if segid is not None:
            idx &= mol.segid == segid
            descr += f' Segid "{segid}"'

        if np.sum(idx) == 0:
            raise RuntimeError(f"No residues found with description ({descr})")

        idx2 = np.where(idx)[0]
        if not np.array_equal(idx2, np.arange(idx2[0], idx2[-1] + 1)):
            raise RuntimeError(
                "Residue with ({}) has non-continuous indexes in PBD file ({})".format(
                    descr, np.where(idx)[0]
                )
            )

        uqins = np.unique(mol.insertion[idx])
        if len(uqins) > 1:
            raise RuntimeError(
                "Residue with ({}) exists with multiple insertions ({}). Define insertion to disambiguate.".format(
                    descr, uqins
                )
            )
        uqchain = np.unique(mol.chain[idx])
        if len(uqchain) > 1:
            raise RuntimeError(
                "Residue with ({}) exists in multiple chains ({}). Define chain to disambiguate.".format(
                    descr, uqchain
                )
            )
        uqseg = np.unique(mol.segid[idx])
        if len(uqseg) > 1:
            raise RuntimeError(
                "Residue with ({}) exists in multiple segments ({}). Define segid to disambiguate.".format(
                    descr, uqseg
                )
            )
        return {
            "resid": resid,
            "insertion": uqins[0],
            "chain": uqchain[0],
            "segid": uqseg[0],
            "idx": idx,
        }

    @staticmethod
    def _findAtom(mol, name, resdict):
        atomidx = resdict["idx"] & (mol.name == name)
        if np.sum(atomidx) == 0:
            raise AtomNotFoundException(
                f'No atoms found with description ({resdict}) and name "{name}".'
            )
        newresdict = {
            "name": name,
            "resid": resdict["resid"],
            "insertion": resdict["insertion"],
            "chain": resdict["chain"],
            "segid": resdict["segid"],
        }
        return newresdict

    @staticmethod
    def _findResname(mol, resdict):
        residx = (
            (mol.resid == resdict["resid"])
            & (mol.insertion == resdict["insertion"])
            & (mol.chain == resdict["chain"])
            & (mol.segid == resdict["segid"])
        )
        uqresname = np.unique(mol.resname[residx])

        if len(uqresname) > 1:
            raise RuntimeError(f"Multiple resnames ({uqresname}) found in ({resdict})")
        return uqresname[0]

    @staticmethod
    def proteinDihedrals(
        mol, sel="protein or resname ACE NME", dih=("psi", "phi"), ff="amber"
    ):
        """Returns a list of tuples containing the four resid/atom pairs for each dihedral of the protein

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        sel : str
            Atom selection string to restrict the atoms for which to calculate dihedrals (e.g. only one of many chains).
            See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__
        dih : tuple
            A tuple of the dihedral types we want to calculate (phi, psi, omega, chi1, chi2, chi3, chi4, chi5)

        Returns
        -------
        dihedrals : list of :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` objects
            A list of Dihedral objects
        """
        mol = mol.copy()
        mol.filter(sel, _logger=False)
        segments = []  # Here I consider segments both chains and segments
        residues = []
        ro = io = co = so = None  # "old" values as in previous atom's
        for r, i, c, s in zip(mol.resid, mol.insertion, mol.chain, mol.segid):
            if co is not None and so is not None and ((c != co) or (s != so)):
                segments.append(residues)
                residues = []
            if (r != ro) or (i != io) or (c != co) or (s != so):
                residues.append((r, i, c, s))
                ro, io, co, so = (r, i, c, s)
        if len(residues) != 0:
            segments.append(residues)

        dihedrals = []
        for s in segments:
            residues = s
            for r in range(len(residues)):
                resid, ins, chain, segid = residues[r]
                starting = r == 0
                ending = r == len(residues) - 1
                if "phi" in dih and not starting:
                    resid2, ins2, _, _ = residues[r - 1]
                    dihedrals.append(
                        Dihedral.phi(mol, resid2, resid, segid, chain, ins2, ins, ff)
                    )
                if "psi" in dih and not ending:
                    resid2, ins2, _, _ = residues[r + 1]
                    dihedrals.append(
                        Dihedral.psi(mol, resid, resid2, segid, chain, ins, ins2, ff)
                    )
                if "omega" in dih and not ending:
                    resid2, ins2, _, _ = residues[r + 1]
                    dihedrals.append(
                        Dihedral.omega(mol, resid, resid2, segid, chain, ins, ins2, ff)
                    )
                if "chi1" in dih:
                    dihedrals.append(Dihedral.chi1(mol, resid, segid, chain, ins, ff))
                if "chi2" in dih:
                    dihedrals.append(Dihedral.chi2(mol, resid, segid, chain, ins, ff))
                if "chi3" in dih:
                    dihedrals.append(Dihedral.chi3(mol, resid, segid, chain, ins, ff))
                if "chi4" in dih:
                    dihedrals.append(Dihedral.chi4(mol, resid, segid, chain, ins, ff))
                if "chi5" in dih:
                    dihedrals.append(Dihedral.chi5(mol, resid, segid, chain, ins, ff))
        return [d for d in dihedrals if d is not None]

    # Sidechain dihedral atoms taken from
    # http://www.ccp14.ac.uk/ccp/web-mirrors/garlic/garlic/commands/dihedrals.html
    @staticmethod
    def phi(
        mol,
        res1,
        res2,
        segid=None,
        chain=None,
        insertion1=None,
        insertion2=None,
        ff="amber",
    ):
        """Constructs a Dihedral object corresponding to the phi angle of res1 and res2

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res1 : int
            The resid of the first residue containing the C atom
        res2 : int
            The resid of the second residue containing the N CA C atoms
        segid : str
            The segment id of the residues
        chain : str
            The chain letter of the residues
        insertion1 : str
            The insertion letter of residue 1
        insertion2 : str
            The insertion letter of residue 2

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        res1dict = Dihedral._findResidue(mol, res1, insertion1, chain, segid)
        res2dict = Dihedral._findResidue(mol, res2, insertion2, chain, segid)

        a1 = Dihedral._findAtom(mol, "C", res1dict)
        try:  # Check if backbone atoms exist. Capped terminals don't have them all
            a2 = Dihedral._findAtom(mol, "N", res2dict)
            a3 = Dihedral._findAtom(mol, "CA", res2dict)
            a4 = Dihedral._findAtom(mol, "C", res2dict)
        except AtomNotFoundException:
            return None

        return Dihedral(a1, a2, a3, a4, dihedraltype="phi", check_valid=False)

    @staticmethod
    def psi(
        mol,
        res1,
        res2,
        segid=None,
        chain=None,
        insertion1=None,
        insertion2=None,
        ff="amber",
    ):
        """Constructs a Dihedral object corresponding to the psi angle of res1 and res2

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res1 : int
            The resid of the first residue containing the N CA C atoms
        res2 : int
            The resid of the second residue containing the N atom
        segid : str
            The segment id of the residues
        chain : str
            The chain letter of the residues
        insertion1 : str
            The insertion letter of residue 1
        insertion2 : str
            The insertion letter of residue 2

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        res1dict = Dihedral._findResidue(mol, res1, insertion1, chain, segid)
        res2dict = Dihedral._findResidue(mol, res2, insertion2, chain, segid)

        try:  # Check if backbone atoms exist. Capped terminals don't have them all
            a1 = Dihedral._findAtom(mol, "N", res1dict)
            a2 = Dihedral._findAtom(mol, "CA", res1dict)
            a3 = Dihedral._findAtom(mol, "C", res1dict)
        except AtomNotFoundException:
            return None

        a4 = Dihedral._findAtom(mol, "N", res2dict)

        return Dihedral(a1, a2, a3, a4, dihedraltype="psi", check_valid=False)

    @staticmethod
    def omega(
        mol,
        res1,
        res2,
        segid=None,
        chain=None,
        insertion1=None,
        insertion2=None,
        ff="amber",
    ):
        """Constructs a Dihedral object corresponding to the omega angle of res1 and res2

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res1 : int
            The resid of the first residue containing the CA C atoms
        res2 : int
            The resid of the second residue containing the N CA atoms
        segid : str
            The segment id of the residues
        chain : str
            The chain letter of the residues
        insertion1 : str
            The insertion letter of residue 1
        insertion2 : str
            The insertion letter of residue 2

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        res1dict = Dihedral._findResidue(mol, res1, insertion1, chain, segid)
        res2dict = Dihedral._findResidue(mol, res2, insertion2, chain, segid)

        try:  # Check if backbone atoms exist. Capped terminals don't have them all
            a1 = Dihedral._findAtom(mol, "CA", res1dict)
            a2 = Dihedral._findAtom(mol, "C", res1dict)
            a3 = Dihedral._findAtom(mol, "N", res2dict)
            a4 = Dihedral._findAtom(mol, "CA", res2dict)
        except AtomNotFoundException:
            return None

        return Dihedral(a1, a2, a3, a4, dihedraltype="omega", check_valid=False)

    @staticmethod
    def chi1(mol, res, segid=None, chain=None, insertion=None, ff="amber"):
        """Constructs a Dihedral object corresponding to the chi1 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi1std = ("N", "CA", "CB", "CG")
        chi1 = {
            "ARG": chi1std,
            "ASN": chi1std,
            "ASP": chi1std,
            "CYS": ("N", "CA", "CB", "SG"),
            "GLN": chi1std,
            "GLU": chi1std,
            "HIS": chi1std,
            "ILE": ("N", "CA", "CB", "CG1"),
            "LEU": chi1std,
            "LYS": chi1std,
            "MET": chi1std,
            "PHE": chi1std,
            "PRO": chi1std,
            "SER": ("N", "CA", "CB", "OG"),
            "THR": ("N", "CA", "CB", "OG1"),
            "TRP": chi1std,
            "TYR": chi1std,
            "VAL": ("N", "CA", "CB", "CG1"),
        }

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        Dihedral._checkKnownResidues(resname)
        if resname not in chi1:
            return None
        return Dihedral(
            Dihedral._findAtom(mol, chi1[resname][0], resdict),
            Dihedral._findAtom(mol, chi1[resname][1], resdict),
            Dihedral._findAtom(mol, chi1[resname][2], resdict),
            Dihedral._findAtom(mol, chi1[resname][3], resdict),
            dihedraltype="chi1",
            check_valid=False,
        )

    @staticmethod
    def chi2(mol, res, segid=None, chain=None, insertion=None, ff="amber"):
        """Constructs a Dihedral object corresponding to the chi2 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        if ff.lower() == "amber":
            chi2std = ("CA", "CB", "CG", "CD")
            chi2 = {
                "ARG": chi2std,
                "ASN": ("CA", "CB", "CG", "OD1"),
                "ASP": ("CA", "CB", "CG", "OD1"),
                "GLN": chi2std,
                "GLU": chi2std,
                "HIS": ("CA", "CB", "CG", "ND1"),
                "ILE": ("CA", "CB", "CG1", "CD1"),
                "LEU": ("CA", "CB", "CG", "CD1"),
                "LYS": chi2std,
                "MET": ("CA", "CB", "CG", "SD"),
                "PHE": ("CA", "CB", "CG", "CD1"),
                "PRO": chi2std,
                "TRP": ("CA", "CB", "CG", "CD1"),
                "TYR": ("CA", "CB", "CG", "CD1"),
            }
        if ff.lower() == "charmm":
            chi2std = ("CA", "CB", "CG", "CD")
            chi2 = {
                "ARG": chi2std,
                "ASN": ("CA", "CB", "CG", "OD1"),
                "ASP": ("CA", "CB", "CG", "OD1"),
                "GLN": chi2std,
                "GLU": chi2std,
                "HIS": ("CA", "CB", "CG", "ND1"),
                "ILE": ("CA", "CB", "CG1", "CD"),
                "LEU": ("CA", "CB", "CG", "CD1"),
                "LYS": chi2std,
                "MET": ("CA", "CB", "CG", "SD"),
                "PHE": ("CA", "CB", "CG", "CD1"),
                "PRO": chi2std,
                "TRP": ("CA", "CB", "CG", "CD1"),
                "TYR": ("CA", "CB", "CG", "CD1"),
            }

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        Dihedral._checkKnownResidues(resname)
        if resname not in chi2:
            return None
        return Dihedral(
            Dihedral._findAtom(mol, chi2[resname][0], resdict),
            Dihedral._findAtom(mol, chi2[resname][1], resdict),
            Dihedral._findAtom(mol, chi2[resname][2], resdict),
            Dihedral._findAtom(mol, chi2[resname][3], resdict),
            dihedraltype="chi2",
            check_valid=False,
        )

    @staticmethod
    def chi3(mol, res, segid=None, chain=None, insertion=None, ff="amber"):
        """Constructs a Dihedral object corresponding to the chi3 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi3 = {
            "ARG": ("CB", "CG", "CD", "NE"),
            "GLN": ("CB", "CG", "CD", "OE1"),
            "GLU": ("CB", "CG", "CD", "OE1"),
            "LYS": ("CB", "CG", "CD", "CE"),
            "MET": ("CB", "CG", "SD", "CE"),
        }

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        Dihedral._checkKnownResidues(resname)
        if resname not in chi3:
            return None
        return Dihedral(
            Dihedral._findAtom(mol, chi3[resname][0], resdict),
            Dihedral._findAtom(mol, chi3[resname][1], resdict),
            Dihedral._findAtom(mol, chi3[resname][2], resdict),
            Dihedral._findAtom(mol, chi3[resname][3], resdict),
            dihedraltype="chi3",
            check_valid=False,
        )

    @staticmethod
    def chi4(mol, res, segid=None, chain=None, insertion=None, ff="amber"):
        """Constructs a Dihedral object corresponding to the chi4 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi4 = {"ARG": ("CG", "CD", "NE", "CZ"), "LYS": ("CG", "CD", "CE", "NZ")}

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        Dihedral._checkKnownResidues(resname)
        if resname not in chi4:
            return None
        return Dihedral(
            Dihedral._findAtom(mol, chi4[resname][0], resdict),
            Dihedral._findAtom(mol, chi4[resname][1], resdict),
            Dihedral._findAtom(mol, chi4[resname][2], resdict),
            Dihedral._findAtom(mol, chi4[resname][3], resdict),
            dihedraltype="chi4",
            check_valid=False,
        )

    @staticmethod
    def chi5(mol, res, segid=None, chain=None, insertion=None, ff="amber"):
        """Constructs a Dihedral object corresponding to the chi5 angle of a residue

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object from which to obtain structural information
        res : int
            The resid of the residue
        segid : str
            The segment id of the residue
        chain : str
            The chain letter of the residue
        insertion : str
            The insertion letter of the residue

        Returns
        -------
        dihedral : :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
            A Dihedral object
        """
        chi5 = {"ARG": ("CD", "NE", "CZ", "NH1")}

        resdict = Dihedral._findResidue(mol, res, insertion, chain, segid)
        resname = Dihedral._findResname(mol, resdict)
        Dihedral._checkKnownResidues(resname)
        if resname not in chi5:
            return None
        return Dihedral(
            Dihedral._findAtom(mol, chi5[resname][0], resdict),
            Dihedral._findAtom(mol, chi5[resname][1], resdict),
            Dihedral._findAtom(mol, chi5[resname][2], resdict),
            Dihedral._findAtom(mol, chi5[resname][3], resdict),
            dihedraltype="chi5",
            check_valid=False,
        )

    @staticmethod
    def _checkKnownResidues(resname):
        knownresnames = [
            "ARG",
            "ASN",
            "ASP",
            "CYS",
            "GLN",
            "GLU",
            "HIS",
            "ILE",
            "LEU",
            "LYS",
            "MET",
            "PHE",
            "PRO",
            "SER",
            "THR",
            "TRP",
            "TYR",
            "VAL",
            "GLY",
            "ALA",
        ]
        if resname not in knownresnames:
            raise RuntimeError(
                "Residue {} not in list of known residues {}. Rename your residues to match these.".format(
                    resname, knownresnames
                )
            )


class MetricDihedral(Projection):
    """Calculates a set of dihedral angles from trajectories

    Parameters
    ----------
    dih : list of :class:`Dihedral <moleculekit.projections.metricdihedral.Dihedral>` object
        You can provide your own list of Dihedral objects. See example.
    sincos : bool, optional
        Set to True to return the dihedral angles as their sine and cosine components. Makes them periodic.
    protsel : str, optional
        Atom selection string for the protein segment for which to calculate dihedral angles. Resids should be unique
        within that segment. See more `here <http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.2/ug/node89.html>`__

    Examples
    --------
    >>> mol = Molecule('3PTB')
    >>> mol.filter('not insertion A')
    >>> met = MetricDihedral()
    >>> met.project(mol)
    >>> # More complicated example
    >>> dih = []
    >>> dih.append(Dihedral.chi1(mol, 45))
    >>> dih.append(Dihedral.psi(mol, 29, 30))
    >>> met = MetricDihedral(dih, protsel='protein and segid 0')
    >>> met.project(mol)
    >>> met.getMapping(mol)
    """

    def __init__(self, dih=None, sincos=True, protsel="protein or resname ACE NME"):
        super().__init__()

        if dih is not None and not isinstance(dih[0], Dihedral):
            raise RuntimeError(
                "Manually passing dihedrals to MetricDihedral requires use of the Dihedral class. Check the example in the documentation"
            )

        self._protsel = protsel
        self._sincos = sincos
        self._dihedrals = dih

    def project(self, mol):
        """Project molecule.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>`
            A :class:`Molecule <moleculekit.molecule.Molecule>` object to project.

        Returns
        -------
        data : np.ndarray
            An array containing the projected data.
        """
        dihedrals = self._getMolProp(mol, "dihedrals")
        return self._calcDihedralAngles(mol, dihedrals, sincos=self._sincos)

    def getMapping(self, mol):
        """Returns the description of each projected dimension.

        Parameters
        ----------
        mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
            A Molecule object which will be used to calculate the descriptions of the projected dimensions.

        Returns
        -------
        map : :class:`DataFrame <pandas.core.frame.DataFrame>` object
            A DataFrame containing the descriptions of each dimension
        """
        dihedrals = self._getMolProp(mol, "dihedrals")

        from pandas import DataFrame

        types = []
        indexes = []
        description = []
        for i, dih in enumerate(dihedrals):
            types += ["dihedral"]
            indexes += [dih]
            mapstr = ""
            for d in dih:
                mapstr += "({} {} {} {} {}) ".format(
                    mol.resname[d],
                    mol.resid[d],
                    mol.name[d],
                    mol.segid[d],
                    mol.chain[d],
                )
            if self._sincos:
                description += ["Sine of angle of " + mapstr]
                description += ["Cosine of angle of " + mapstr]
                types += ["dihedral"]
                indexes += [dih]
            else:
                description += ["Angle of " + mapstr]
        return DataFrame(
            {"type": types, "atomIndexes": indexes, "description": description}
        )

    def _calculateMolProp(self, mol, props="all"):
        from moleculekit.util import ensurelist

        res = {}
        protsel = mol.atomselect(self._protsel)

        if self._dihedrals is None:  # Default phi psi dihedrals
            dihedrals = Dihedral.proteinDihedrals(mol, protsel)
        else:
            dihedrals = ensurelist(self._dihedrals)

        res["dihedrals"] = Dihedral.dihedralsToIndexes(mol, dihedrals, protsel)
        return res

    def _calcDihedralAngles(self, mol, dihedrals, sincos=True):
        from moleculekit.dihedral import dihedralAngle

        metric = np.zeros((np.size(mol.coords, 2), len(dihedrals)))

        for i, dih in enumerate(dihedrals):
            metric[:, i] = np.rad2deg(dihedralAngle(mol.coords[dih, :, :]))

        if sincos:
            sc_metric = np.zeros((np.size(metric, 0), np.size(metric, 1) * 2))
            sc_metric[:, 0::2] = np.sin(metric * np.pi / 180.0)
            sc_metric[:, 1::2] = np.cos(metric * np.pi / 180.0)
            metric = sc_metric
        return metric.astype(np.float32)


import unittest


class _TestMetricDihedral(unittest.TestCase):
    def test_dihedral_traj(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(
            path.join(home(dataDir="test-projections"), "trajectory", "filtered.pdb")
        )
        mol.read(path.join(home(dataDir="test-projections"), "trajectory", "traj.xtc"))

        metr = MetricDihedral(protsel="protein")
        data = metr.project(mol)
        dataref = np.load(
            path.join(home(dataDir="test-projections"), "metricdihedral", "ref.npy")
        )
        assert np.allclose(
            data, dataref, atol=1e-03
        ), "Diherdals calculation gave different results from reference"

    def test_dihedral_5mat(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from moleculekit.tools.autosegment import autoSegment
        from os import path

        mol = Molecule("5MAT")
        mol.filter("not insertion A and not altloc A B", _logger=False)
        mol = autoSegment(mol, _logger=False)

        data = MetricDihedral().project(mol)
        dataref = np.load(
            path.join(home(dataDir="test-projections"), "metricdihedral", "5mat.npy")
        )
        assert np.allclose(
            data, dataref, atol=1e-03
        ), "Diherdals calculation gave different results from reference"

        ref_idx = np.load(
            path.join(
                home(dataDir="test-projections"),
                "metricdihedral",
                "5mat_mapping_indexes.npy",
            )
        )
        mapping = MetricDihedral().getMapping(mol)
        mapping_idx = np.vstack(mapping.atomIndexes.to_numpy())

        assert np.array_equal(
            mapping_idx, ref_idx
        ), "Mapping of atom indexes has changed"

    def test_dialanine_ace_nme(self):
        from moleculekit.molecule import Molecule
        from moleculekit.home import home
        from os import path

        mol = Molecule(
            path.join(
                home(dataDir="test-projections"),
                "metricdihedral",
                "dialanine-peptide.pdb",
            )
        )
        data = MetricDihedral().project(mol)

        refarray = np.array(
            [[-0.71247578, -0.70169669, 0.27399951, -0.96172982]], dtype=np.float32
        )
        assert np.allclose(refarray, data)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import logging
import unittest
import tempfile
import numpy as np
import pandas as pd
import os
from moleculekit.molecule import Molecule, mol_equal, UniqueResidueID


logger = logging.getLogger(__name__)


def _selToHoldList(mol, sel):
    ret = None
    if sel:
        sel = mol.atomselect(sel)
        ret = [
            list(x)
            for x in set(
                tuple(x)
                for x in zip(mol.resid[sel], mol.chain[sel], mol.insertion[sel])
            )
        ]
    return ret


def _fixup_water_names(mol):
    """Rename WAT / OW HW HW atoms as O H1 H2"""
    mol.set("name", "O", sel="resname WAT and name OW")
    mol.set("name", "H1", sel="resname WAT and name HW and serial % 2 == 0")
    mol.set("name", "H2", sel="resname WAT and name HW and serial % 2 == 1")


def _warn_if_contains_DUM(mol):
    """Warn if any DUM atom is there"""
    if any(mol.atomselect("resname DUM")):
        logger.warning(
            "OPM's DUM residues must be filtered out before preparation. Continuing, but crash likely."
        )


def _check_chain_and_segid(mol, verbose):
    from moleculekit.util import sequenceID
    import string

    emptychains = mol.chain == ""
    emptysegids = mol.segid == ""

    if np.all(emptychains) and np.all(emptysegids):
        raise RuntimeError(
            "No chains or segments defined in Molecule.chain / Molecule.segid. Please assign either to continue with preparation."
        )

    if np.all(emptychains) and np.any(~emptysegids):
        logger.info(
            "No chains defined in Molecule. Using segment IDs as chains for protein preparation."
        )
        mol = mol.copy()
        chain_alphabet = np.array(
            list(string.digits + string.ascii_uppercase + string.ascii_lowercase)
        )
        mol.chain[:] = chain_alphabet[sequenceID(mol.segid)]

    if np.any(~emptysegids) and np.any(~emptychains):
        chainseq = sequenceID(mol.chain)
        segidseq = sequenceID(mol.segid)
        if not np.array_equal(chainseq, segidseq):
            logger.warning(
                "Both chains and segments are defined in Molecule.chain / Molecule.segid, however they are inconsistent. "
                "Protein preparation will use the chain information."
            )

    if verbose:
        chainids = np.unique(mol.chain)
        if np.any([len(cc) > 1 for cc in chainids]):
            raise RuntimeError(
                "The chain field should only contain a single character."
            )

        print("\n---- Molecule chain report ----")
        for c in chainids:
            chainatoms = np.where(mol.chain == c)[0]
            firstatom = chainatoms[0]
            lastatom = chainatoms[-1]
            print(f"Chain {c}:")
            print(
                f"    First residue: {mol.resname[firstatom]}:{mol.resid[firstatom]}:{mol.insertion[firstatom]}"
            )
            print(
                f"    Final residue: {mol.resname[lastatom]}:{mol.resid[lastatom]}:{mol.insertion[lastatom]}"
            )
        print("---- End of chain report ----\n")

    return mol


def _generate_nonstandard_residues_ff(mol, definition, forcefield, _molkit_ff=True):
    import tempfile
    from moleculekit.tools.preparation_customres import _get_custom_ff
    from moleculekit.tools.preparation_customres import (
        _generate_custom_residue,
        _mol_to_dat_def,
        _mol_to_xml_def,
    )

    uqprot_resn = np.unique(mol.get("resname", sel="protein"))
    not_in_ff = []
    for resn in uqprot_resn:
        if not forcefield.has_residue(resn):
            not_in_ff.append(resn)

    if len(not_in_ff) == 0:
        return definition, forcefield

    try:
        import aceprep
    except ImportError:
        raise RuntimeError(
            "To protonate non-standard aminoacids you need the aceprep library. Please contact Acellera info@acellera.com for more information."
        )

    with tempfile.TemporaryDirectory() as tmpdir:
        for res in not_in_ff:
            try:
                logger.warning(f"Attempting to template non-standard residue {res}...")
                molc = mol.copy()
                # Hacky way of getting the first molecule, if there are copies
                molresn = molc.resname == res
                firstname = mol.name[molresn][0]
                lastname = mol.name[molresn][-1]
                start = np.where(molresn & (mol.name == firstname))[0][0]
                end = np.where(molresn & (mol.name == lastname))[0][0]
                # Remove all other stuff
                molc.filter(f"index {start} to {end}", _logger=False)
                cres = _generate_custom_residue(molc, res)
                _mol_to_xml_def(cres, os.path.join(tmpdir, f"{res}.xml"))
                _mol_to_dat_def(cres, os.path.join(tmpdir, f"{res}.dat"))
                logger.warning(f"Succesfully templated non-standard residue {res}.")
            except Exception as e:
                raise RuntimeError(
                    f"Failed to protonate non-standard residue {res}. Please remove it from the protein or mutate it to continue preparation. Detailed error message: {e}"
                )
        definition, forcefield = _get_custom_ff(user_ff=tmpdir, molkit_ff=_molkit_ff)
    return definition, forcefield


def _detect_nonpeptidic_bonds(mol):
    coordination_ions = ("Ca", "Zn", "Mn", "Cu", "Fe", "Ni", "Mo", "Re", "Na")

    prot_idx = mol.atomselect("protein", indexes=True)
    ion_idx = np.where(np.isin(mol.element, coordination_ions))[0]
    same_resname = np.isin(
        mol.resname[ion_idx], [cc.upper() for cc in coordination_ions]
    )
    ion_idx = ion_idx[same_resname]

    # Bonds where only 1 atom belongs to protein
    bond_mask = np.isin(mol.bonds, prot_idx)
    # Exclude ions
    bond_mask[np.any(np.isin(mol.bonds, ion_idx), axis=1), :] = False
    inter_bonds = np.sum(bond_mask, axis=1) == 1

    bonds = mol.bonds[inter_bonds, :]
    b_prot_idx = bonds[bond_mask[inter_bonds]]
    b_other_idx = bonds[~bond_mask[inter_bonds]]

    if not len(b_prot_idx):
        return []

    logger.info(
        f"Found {len(b_prot_idx)} covalent bonds from protein to non-protein molecules."
    )
    return np.vstack([b_prot_idx, b_other_idx]).T


def _delete_no_titrate(pka_list, no_titr):
    pkas = []
    for res in pka_list:
        key = (res["res_num"], res["chain_id"].strip(), res["ins_code"].strip())
        if key not in no_titr:
            pkas.append(res)
        else:
            logger.info(
                f"Skipped titration of residue {res['res_name']}:{key[1]}:{key[0]}{key[2]}"
            )
    return pkas


def _apply_patches_force_prot(biomolecule, force_protonation):
    for res in biomolecule.residues:
        key = (res.res_seq, res.chain_id, res.ins_code)
        if key in force_protonation:
            logger.info(
                f"Forcing protonation of residue {res.name}:{key[1]}:{key[0]}{key[2]} to {force_protonation[key]}"
            )
            biomolecule.apply_patch(force_protonation[key], res)


def _pdb2pqr(
    pdb_file,
    definition,
    forcefield_,
    ph=7.4,
    assign_only=False,
    clean=False,
    debump=True,
    opt=True,
    drop_water=False,
    ligand=None,
    titrate=True,
    neutraln=False,
    neutralc=False,
    no_prot=None,
    no_opt=None,
    no_titr=None,
    force_protonation=None,
    propka_args=None,
):
    from pdb2pqr.io import get_molecule
    from pdb2pqr import forcefield, hydrogens
    from pdb2pqr.debump import Debump
    from pdb2pqr.main import is_repairable, run_propka, setup_molecule
    from pdb2pqr.main import drop_water as drop_water_func
    from propka.lib import build_parser

    propka_parser = build_parser()
    propka_parser.add_argument("--keep-chain", action="store_true", default=False)

    if propka_args is None:
        propka_args = {}
    propka_args_list = []
    for key, value in propka_args.items():
        propka_args_list += [key, value]
    propka_args_list += ["--log-level", "WARNING", "--pH", str(ph), "STEFAN"]
    propka_args = propka_parser.parse_args(propka_args_list)

    if assign_only or clean:
        debump = False
        opt = False

    pdblist, _ = get_molecule(pdb_file)
    if drop_water:
        pdblist = drop_water_func(pdblist)

    biomolecule, definition, ligand = setup_molecule(pdblist, definition, ligand)
    biomolecule.set_termini(neutraln, neutralc)
    biomolecule.update_bonds()
    if clean:
        return None, None, biomolecule

    hydrogen_handler = hydrogens.create_handler()
    debumper = Debump(biomolecule)
    pkas = None
    if assign_only:
        # TODO - I don't understand why HIS needs to be set to HIP for
        # assign-only
        biomolecule.set_hip()
    else:
        if is_repairable(biomolecule, ligand is not None):
            biomolecule.repair_heavy()

        biomolecule.update_ss_bridges()
        if debump:
            try:
                debumper.debump_biomolecule()
            except ValueError as err:
                err = f"Unable to debump biomolecule. {err}"
                raise ValueError(err)
        if titrate:
            biomolecule.remove_hydrogens()
            pka_list, _ = run_propka(propka_args, biomolecule)
            # logger.info(f"PROPKA information:\n{pka_str}")

            # STEFAN mod: Delete pka values for residues we should not titrate
            pkas = _delete_no_titrate(pka_list, no_titr)
            # STEFAN mod: Force specific protonation states to residues
            _apply_patches_force_prot(biomolecule, force_protonation)

            biomolecule.apply_pka_values(
                forcefield_.name,
                ph,
                {row["group_label"]: row["pKa"] for row in pkas},
            )

        # TODO: Remove this try/except once pdb2pqr makes new release
        try:
            biomolecule.add_hydrogens(no_prot)  # STEFAN mod: Don't protonate residues
        except Exception:
            logger.error(
                "Disabling residue protonation requires pdb2pqr version >=3.3.1. All residues will be protonated"
            )
            biomolecule.add_hydrogens()

        if debump:
            debumper.debump_biomolecule()

        hydrogen_routines = hydrogens.HydrogenRoutines(debumper, hydrogen_handler)
        if opt:
            hydrogen_routines.set_optimizeable_hydrogens()
            biomolecule.hold_residues(no_opt)  # STEFAN mod: Don't optimize residues
            hydrogen_routines.initialize_full_optimization()
        else:
            hydrogen_routines.initialize_wat_optimization()
        hydrogen_routines.optimize_hydrogens()
        hydrogen_routines.cleanup()

    biomolecule.set_states()
    matched_atoms, missing_atoms = biomolecule.apply_force_field(forcefield_)
    if ligand is not None:
        logger.warning("Using ZAP9 forcefield for ligand radii.")
        ligand.assign_parameters()
        lig_atoms = []
        for residue in biomolecule.residues:
            tot_charge = 0
            for pdb_atom in residue.atoms:
                # Only check residues with HETATM
                if pdb_atom.type == "ATOM":
                    break
                try:
                    mol2_atom = ligand.atoms[pdb_atom.name]
                    pdb_atom.radius = mol2_atom.radius
                    pdb_atom.ffcharge = mol2_atom.charge
                    tot_charge += mol2_atom.charge
                    lig_atoms.append(pdb_atom)
                except KeyError:
                    err = f"Can't find HETATM {residue.name} {residue.res_seq} {pdb_atom.name} in MOL2 file"
                    logger.warning(err)
                    missing_atoms.append(pdb_atom)
        matched_atoms += lig_atoms

    name_scheme = forcefield.Forcefield("amber", definition, None)
    biomolecule.apply_name_scheme(name_scheme)

    return missing_atoms, pkas, biomolecule


def _atomsel_to_hold(mol_in, sel):
    idx = mol_in.atomselect(sel, indexes=True)
    if len(idx) == 0:
        raise RuntimeError(f"Could not select any atoms with atomselection {sel}")
    residues = [(mol_in.resid[i], mol_in.chain[i], mol_in.insertion[i]) for i in idx]
    residues = list(set(residues))
    if len(residues) != 1:
        raise RuntimeError(
            f"no_opt selection {sel} selected multiple residues {residues}. Please be more specific on which residues to hold."
        )
    return residues[0]


def _get_hold_residues(
    mol_in, no_opt, no_prot, no_titr, force_protonation, nonpeptidic_bonds
):
    # Converts atom selections used for no optimization residues and no protonation residues
    # to tuples of (resid, chain, insertion) which are used by pdb2pqr as unique residue identifiers
    _no_opt = []
    if no_opt is not None:
        for sel in no_opt:
            _no_opt.append(_atomsel_to_hold(mol_in, sel))

    _no_prot = []
    if no_prot is not None:
        for sel in no_prot:
            _no_prot.append(_atomsel_to_hold(mol_in, sel))

    _no_titr = []
    if no_titr is not None:
        for sel in no_titr:
            _no_titr.append(_atomsel_to_hold(mol_in, sel))

    _force_prot = {}
    if force_protonation is not None:
        for sel, resn in force_protonation:
            _force_prot[_atomsel_to_hold(mol_in, sel)] = resn

    # Add residues which should not be protonated to the residues which should not be titrated
    _no_titr += _no_prot
    # Add residues which have forced protonations to the residues which should not be titrated
    _no_titr += list(_force_prot.keys())

    if len(nonpeptidic_bonds) != 0:
        for nn in nonpeptidic_bonds:
            r1 = UniqueResidueID.fromMolecule(mol_in, idx=nn[0])
            r1 = f"{r1.resname}:{r1.chain}:{r1.resid}{r1.insertion}"
            r2 = UniqueResidueID.fromMolecule(mol_in, idx=nn[1])
            r2 = f"{r2.resname}:{r2.chain}:{r2.resid}{r2.insertion}"
            logger.info(
                f"Freezing protein residue {r1} bonded to non-protein molecule {r2}"
            )
            val = [
                (mol_in.resid[nn[0]], mol_in.chain[nn[0]], mol_in.insertion[nn[0]]),
                (mol_in.resid[nn[1]], mol_in.chain[nn[1]], mol_in.insertion[nn[1]]),
            ]
            _no_opt += val
            _no_titr += val
            if val[0] not in _force_prot:
                # Only disable protonation if there is no force-protonation
                _no_prot += val

    return _no_opt, _no_prot, _no_titr, _force_prot


def _check_frozen_histidines(mol_in, _no_prot):
    # Histidines require protonation by pdb2pqr. We cannot freeze them
    frozen_his = []
    for key in _no_prot:
        sel = (
            (mol_in.resid == key[0])
            & (mol_in.chain == key[1])
            & (mol_in.insertion == key[2])
        )
        if mol_in.resname[sel][0] == "HIS":
            frozen_his.append(key)

    if len(frozen_his):
        res = ", ".join([f"HIS:{key[1]}:{key[0]}{key[2]}" for key in frozen_his])
        raise RuntimeError(
            f"Histidines {res} are not auto-titrated. You need to manually define protonation states for them using the force_protonation argument."
        )


def _prepare_nucleics(mol):
    # Renames residues to the names expected by PARSE. Fixes issues with 5' G atoms.
    nucl_mask = mol.atomselect("nucleic")
    uq_resn = mol.resname[nucl_mask]

    # Clean 5' G problematic atoms
    for ch in np.unique(mol.get("chain", "nucleic")):
        resid = mol.resid[(mol.chain == ch) & mol.atomselect("nucleic")][0]
        mol.remove(
            f"chain {ch} and resid {resid} and (resname G) and (name P OP1 OP2)",
            _logger=False,
        )

    mapping = {
        "T": "DT",
        "U": "RU",
        "G": "RG",
        "C": "RC",
        "A": "RA",
        "DG": "DG",
        "DC": "DC",
        "DA": "DA",
        "DT": "DT",
    }
    for res in uq_resn:
        mol.resname[mol.resname == res] = mapping[res]


def proteinPrepare(
    mol_in,
    pH=7.0,
    verbose=0,
    returnDetails=False,
    hydrophobicThickness=None,
    holdSelection=None,
    _loggerLevel=None,
):
    logger.warning(
        "proteinPrepare has been deprecated in favor of the systemPrepare function and will soon be removed. "
        "Please look at the documentation of systemPrepare for more information."
    )
    return systemPrepare(
        mol_in,
        pH=pH,
        verbose=verbose,
        return_details=returnDetails,
        hydrophobic_thickness=hydrophobicThickness,
        _logger_level=_loggerLevel,
    )


def systemPrepare(
    mol_in,
    titration=True,
    pH=7.4,
    force_protonation=None,
    no_opt=None,
    no_prot=None,
    no_titr=None,
    hold_nonpeptidic_bonds=True,
    verbose=True,
    return_details=False,
    hydrophobic_thickness=None,
    plot_pka=None,
    _logger_level="ERROR",
    _molkit_ff=True,
):
    """Prepare molecular systems through protonation and h-bond optimization.

    The preparation routine protonates and optimizes protein and nucleic residues.
    It will also take into account any non-protein, non-nucleic molecules for the pKa calculation
    but will not attempt to protonate or optimize those.

    Returns a Molecule object, where residues have been renamed to follow
    internal conventions on protonation (below). Coordinates are changed to
    optimize the H-bonding network.

    The following residue names are used in the returned molecule:

    === ===============================
    ASH Neutral ASP
    CYX SS-bonded CYS
    CYM Negative CYS
    GLH Neutral GLU
    HIP Positive HIS
    HID Neutral HIS, proton HD1 present
    HIE Neutral HIS, proton HE2 present
    LYN Neutral LYS
    TYM Negative TYR
    AR0 Neutral ARG
    === ===============================

    ========= ======= =========
    Charge +1 Neutral Charge -1
    ========= ======= =========
    -         ASH     ASP
    -         CYS     CYM
    -         GLH     GLU
    HIP       HID/HIE -
    LYS       LYN     -
    -         TYR     TYM
    ARG       AR0     -
    ========= ======= =========

    A detailed table about the residues modified is returned (as a second return value) when
    return_details is True .

    If hydrophobic_thickness is set to a positive value 2*h, a warning is produced for titratable residues
    having -h<z<h and are buried in the protein by less than 75%. Note that the heuristic for the
    detection of membrane-exposed residues is very crude; the "buried fraction" computation
    (from propKa) is approximate; also, in the presence of cavities,
    residues may be solvent-exposed independently from their z location.


    Notes
    -----

    Features:
     - assigns protonation states via propKa
     - flips residues to optimize H-bonding network
     - debumps collisions
     - fills in missing atoms, e.g. hydrogen atoms


    Parameters
    ----------
    mol_in : moleculekit.molecule.Molecule
        the object to be optimized
    titration : bool
        If set to True it will use propka to set the titration state of residues. Otherwise it will just add and optimize the hydrogens.
    pH : float
        pH to decide titration
    verbose : bool
        verbosity
    return_details : bool
        whether to return just the prepared Molecule (False, default) or a molecule *and* a ResidueInfo
        object including computed properties
    hydrophobic_thickness : float
        the thickness of the membrane in which the protein is embedded, or None if globular protein.
        Used to provide a warning about membrane-exposed residues.
    force_protonation : list of tuples
        Allows the user to force specific protonations on residues. This can be done by providing a list of tuples,
        one for each residue we want to force. i.e. [("protein and resid 40", "HID")] will force the protonation of
        the first atomselection to the second resname. Atomselections should be valid VMD atomselections.
    no_opt : list of str
        Allows the user to disable optimization for specific residues. For example if the user determines that a
        residue flip or change of coordinates performed by this method causes issues in the structure, they can
        disable optimization on that residue by passing an atomselection for the residue to hold. i.e. ["protein and resid 23"].
    no_prot : list of str
        Same as no_opt but disables the addition of hydrogens to specific residues.
    no_titr : list of str
        Same as no_opt but disables the titration of specific residues.
    hold_nonpeptidic_bonds : bool
        When set to True, systemPrepare will automatically not optimize, protonate or titrate protein residues
        which are covalently bound to non-protein molecules. When set to False, systemPrepare will optimize them
        ignoring the covalent bond, meaning it may break the bonds or add hydrogen atoms between the bonds.
    plot_pka : str
        Provide a file path with .png extension to draw the titration diagram for the system residues.

    Returns
    -------
    mol_out : moleculekit.molecule.Molecule
        the molecule titrated and optimized. The molecule object contains an additional attribute,
    details : pandas.DataFrame
        A table of residues with the corresponding protonation states, pKas, and other information


    Examples
    --------
    >>> tryp = Molecule('3PTB')
    >>> tryp_op, df = systemPrepare(tryp, return_details=True)
    >>> tryp_op.write('/tmp/3PTB_prepared.pdb')
    >>> df.to_excel("/tmp/tryp-report.csv")
    >>> df                                                        # doctest: +NORMALIZE_WHITESPACE
    resname protonation  resid insertion chain segid       pKa    buried
    0       ILE         ILE     16               A     0  7.413075  0.839286
    1       VAL         VAL     17               A     0       NaN       NaN
    2       GLY         GLY     18               A     0       NaN       NaN
    3       GLY         GLY     19               A     0       NaN       NaN
    4       TYR         TYR     20               A     0  9.590845  0.146429
    ..      ...         ...    ...       ...   ...   ...       ...       ...
    282     HOH         WAT    804               A     1       NaN       NaN
    283     HOH         WAT    805               A     1       NaN       NaN
    284     HOH         WAT    807               A     1       NaN       NaN
    285     HOH         WAT    808               A     1       NaN       NaN
    286     HOH         WAT    809               A     1       NaN       NaN

    [287 rows x 8 columns]

    >>> tryp_op = systemPrepare(tryp, pH=1.0)
    >>> tryp_op.write('/tmp/3PTB_pH1.pdb')

    The following will force the preparation to freeze residues 36 and 49 in place
    >>> tryp_op = systemPrepare(tryp, no_opt=["protein and resid 36", "chain A and resid 49"])

    The following will disable protonation on residue 32 of the protein
    >>> tryp_op = systemPrepare(tryp, no_prot=["protein and resid 32",])

    The following will disable titration and protonation on residue 32
    >>> tryp_op = systemPrepare(tryp, no_titr=["protein and resid 32",], no_prot=["protein and resid 32",])

    The following will force residue 40 protonation to HIE and 57 to HIP
    >>> tryp_op = systemPrepare(tryp, force_protonation=[("protein and resid 40", "HIE"), ("protein and resid 57", "HIP")])
    """
    try:
        from pdb2pqr.config import VERSION
    except ImportError:
        raise ImportError(
            "pdb2pqr not installed. To use the system preparation features please do `conda install pdb2pqr -c conda-forge`"
        )
    from moleculekit.tools.preparation_customres import _get_custom_ff

    old_level = logger.getEffectiveLevel()
    if not verbose:
        logger.setLevel(logging.WARNING)

    if _logger_level is not None:
        # logger.setLevel(_loggerLevel)
        logging.getLogger(f"PDB2PQR{VERSION}").setLevel(_logger_level)
        logging.getLogger(f"PDB2PQR{VERSION}").propagate = False
        logging.getLogger("pdb2pqr").setLevel(_logger_level)
        logging.getLogger("propka").setLevel(_logger_level)
    logger.debug("Starting.")

    mol_in = mol_in.copy()

    _warn_if_contains_DUM(mol_in)

    mol_in = _check_chain_and_segid(mol_in, verbose)

    _prepare_nucleics(mol_in)

    definition, forcefield = _get_custom_ff(molkit_ff=_molkit_ff)
    definition, forcefield = _generate_nonstandard_residues_ff(
        mol_in, definition, forcefield, _molkit_ff
    )
    nonpept = None
    if hold_nonpeptidic_bonds:
        nonpept = _detect_nonpeptidic_bonds(mol_in)

    _no_opt, _no_prot, _no_titr, _force_prot = _get_hold_residues(
        mol_in, no_opt, no_prot, no_titr, force_protonation, nonpept
    )
    _check_frozen_histidines(mol_in, _no_prot)

    with tempfile.TemporaryDirectory() as tmpdir:
        tmpin = os.path.join(tmpdir, "input.pdb")
        mol_in.write(tmpin)

        missedres, pka_df, biomolecule = _pdb2pqr(
            tmpin,
            ph=pH,
            titrate=titration,
            no_opt=_no_opt,
            no_prot=_no_prot,
            no_titr=_no_titr,
            force_protonation=_force_prot,
            definition=definition,
            forcefield_=forcefield,
        )
        mol_out = _biomolecule_to_molecule(biomolecule)

    # Diagnostics
    missedres = set([m.residue.name for m in missedres])
    if len(missedres):
        logger.warning(
            f"The following residues have not been optimized: {', '.join(missedres)}"
        )

    mol_out.box = mol_in.box
    _fixup_water_names(mol_out)

    df = _create_table(mol_in, mol_out, pka_df)

    _list_modifications(df)

    if titration:
        if plot_pka is not None:
            _get_pka_plot(df, plot_pka)
        _warn_pk_close_to_ph(df, pH)
        if hydrophobic_thickness:
            # TODO: I think this only works if the protein is assumed aligned to the membrane and the membrane is centered at Z=0
            _warn_buried_residues(df, mol_out, hydrophobic_thickness)

    logger.setLevel(old_level)

    if return_details:
        return mol_out, df
    return mol_out


def _biomolecule_to_molecule(biomolecule):
    propmap = [
        ("type", "record"),
        ("serial", "serial"),
        ("name", "name"),
        ("alt_loc", "altloc"),
        ("res_name", "resname"),
        ("chain_id", "chain"),
        ("res_seq", "resid"),
        ("ins_code", "insertion"),
        ("occupancy", "occupancy"),
        ("temp_factor", "beta"),
        ("seg_id", "segid"),
        ("element", "element"),
        ("charge", "charge"),
    ]
    mol = Molecule().empty(len(biomolecule.atoms))
    mol.coords = np.zeros((mol.numAtoms, 3, 1), dtype=Molecule._dtypes["coords"])
    bonds = []
    for i, atom in enumerate(biomolecule.atoms):
        for pp1, pp2 in propmap:
            val = getattr(atom, pp1)
            if pp1 == "charge":
                val = float(val) if val != "" else 0
            if pp1 == "element":
                val = str(val).capitalize()
            mol.__dict__[pp2][i] = val
        mol.coords[i, :, 0] = [atom.x, atom.y, atom.z]
        bonds += [[i, x.serial - 1] for x in atom.bonds]

    # guessedBonds = mol._guessBonds()
    # bb = mol.atomselect("backbone", indexes=True)
    # bb = guessedBonds[np.all(np.isin(guessedBonds, bb), axis=1)]
    # mol.bonds = np.vstack(bb.tolist() + bonds)

    return mol


_table_dtypes = {
    "resname": str,
    "protonation": str,
    "resid": int,
    "insertion": str,
    "chain": str,
    "segid": str,
    "pKa": float,
    "buried": float,
}


def _create_table(mol_in, mol_out, pka_df):
    import pandas as pd

    uq_tuple = []
    for e in zip(
        mol_in.resname, mol_in.resid, mol_in.insertion, mol_in.chain, mol_in.segid
    ):
        if e not in uq_tuple:
            uq_tuple.append(e)

    data = []
    for tup in uq_tuple:
        old_resn = tup[0]
        new_mask = (
            (mol_out.resid == tup[1])
            & (mol_out.insertion == tup[2])
            & (mol_out.chain == tup[3])
            & (mol_out.segid == tup[4])
        )
        if not np.any(new_mask):
            raise RuntimeError(
                f"Unable to find residue {' '.join(tup)} after preparation"
            )
        new_idx = np.where(new_mask)[0][0]
        resname = mol_out.resname[new_idx]
        resid = mol_out.resid[new_idx]
        insertion = mol_out.insertion[new_idx]
        chain = mol_out.chain[new_idx]
        segid = mol_out.segid[new_idx]
        curr_data = [old_resn, resname, resid, insertion, chain, segid]
        if pka_df is not None:
            found = False
            for propkadata in pka_df:
                if (
                    propkadata["res_num"] == resid
                    and propkadata["ins_code"].strip() == insertion.strip()
                    and propkadata["chain_id"] == chain
                ):
                    curr_data += [propkadata["pKa"], propkadata["buried"]]
                    found = True
                    break
            if not found:
                curr_data += [np.nan, np.nan]
        data.append(curr_data)

    cols = [
        "resname",
        "protonation",
        "resid",
        "insertion",
        "chain",
        "segid",
    ]
    dtypes = _table_dtypes
    if pka_df is not None:
        cols += ["pKa", "buried"]
    else:
        del dtypes["pKa"]
        del dtypes["buried"]
    df = pd.DataFrame(data=data, columns=cols)
    df = df.astype(dtypes)
    return df


def _list_modifications(df):
    for _, row in df[df.resname != df.protonation].iterrows():
        if row.resname in ["HOH", "WAT"]:
            continue

        old_resn = row.resname
        new_resn = row.protonation
        ch = row.chain
        rid = row.resid
        ins = row.insertion.strip()
        logger.info(f"Modified residue {old_resn}:{ch}:{rid}{ins} to {new_resn}")


def _warn_pk_close_to_ph(df, pH, tol=1.0):
    dubious = df[abs(df.pKa - pH) < tol]
    if len(dubious):
        logger.warning(
            f"Dubious protonation state: the pKa of {len(dubious)} residues is within {tol:.1f} units of pH {pH:.1f}."
        )
        for _, dr in dubious.iterrows():
            logger.warning(
                f"Dubious protonation state:    {dr.resname:4s} {dr.resid:>4d}{dr.insertion.strip()} {dr.chain:1s} (pKa={dr.pKa:5.2f})"
            )


def _warn_buried_residues(df, mol_out, hydrophobic_thickness, maxBuried=0.75):
    ht = hydrophobic_thickness / 2.0
    count = 0
    for _, row in df[df.buried > maxBuried].iterrows():
        sele = (
            (mol_out.chain == row.chain)
            & (mol_out.resid == row.resid)
            & (mol_out.insertion == row.insertion)
        )
        mean_z = np.mean(mol_out.coords[sele, 2, 0])
        if -ht < mean_z < ht:
            count += 1

    if count:
        logger.warning(
            f"Predictions for {count} residues may be incorrect because they are "
            + f"exposed to the membrane ({-ht:.1f}<z<{ht:.2f} and buried<{maxBuried:.1f}%)."
        )


def _get_pka_plot(df, outname, pH=7.4, figSizeX=13, dpk=1.0, font_size=12):
    """Internal function to build the protonation diagram"""
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib.lines import Line2D
    from matplotlib.colors import LinearSegmentedColormap
    import matplotlib.patheffects as PathEffects

    # Shading
    Xe = np.array([[1, 0], [1, 0]])

    # Shading colors http://matplotlib.org/examples/pylab_examples/custom_cmap.html
    neutral_grey = (0.7, 0.7, 0.7)
    my_red = (0.98, 0.41, 0.29)
    my_blue = (0.42, 0.68, 0.84)
    grey_red = LinearSegmentedColormap.from_list("grey_red", [neutral_grey, my_red])
    grey_blue = LinearSegmentedColormap.from_list("grey_blue", [neutral_grey, my_blue])
    eps = 0.01  # Tiny overprint to avoid very thin white lines
    outline = [PathEffects.withStroke(linewidth=2, foreground="w")]

    # Color for pk values
    pKa_color = "black"
    pKa_fontsize = 8
    dtxt = 0  # Displacement

    # Or we could change the figure size, which scales axes
    # http://stackoverflow.com/questions/3899980/how-to-change-the-font-size-on-a-matplotlib-plot
    plt.rc("font", family="monospace")
    plt.rc("font", size=font_size)  # controls default text sizes
    plt.rc("axes", titlesize=font_size)  # fontsize of the axes title
    plt.rc("axes", labelsize=font_size)  # fontsize of the x and y labels
    plt.rc("xtick", labelsize=font_size)  # fontsize of the tick labels
    plt.rc("ytick", labelsize=font_size)  # fontsize of the tick labels
    plt.rc("legend", fontsize=font_size)  # legend fontsize
    plt.rc("figure", titlesize=font_size)  # fontsize of the figure title

    # Constants
    acidicResidues = ["ASP", "GLU", "TYR", "C-"]

    df = df[df.pKa < 99]

    # Format residue labels
    labels = []
    pKas = []
    restypes = []
    for _, row in df.iterrows():
        dub = "(!)" if abs(row.pKa - pH) < dpk else ""
        labels.append(
            f"{dub} {row.chain}:{row.resid}{row.insertion.strip()}-{row.resname}"
        )
        pKas.append(row.pKa)
        restypes.append("neg" if row.resname in acidicResidues else "pos")

    N = len(df)
    xmin, xmax = xlim = 0, 14
    ymin, ymax = ylim = -1, N

    width = 0.8  # Of each band

    # So, arbitrarily, 40 residues are square
    sizePerBand = figSizeX * (N / 40)
    figsize = (figSizeX, sizePerBand)
    fig = plt.figure(figsize=figsize)
    ax = fig.add_subplot(111, xlim=xlim, ylim=ylim, autoscale_on=False)

    ax.xaxis.tick_top()
    ax.set_xlabel("pKa")
    ax.xaxis.set_label_position("top")

    ax.yaxis.set_ticks(range(N))
    ax.yaxis.set_ticklabels(labels)
    ax.invert_yaxis()

    for i in range(N):
        left = xmin
        right = xmax
        top = i + width / 2
        bottom = i - width / 2
        pk = pKas[i]
        restype = restypes[i]

        if restype == "neg":
            ax.imshow(
                Xe * 0,
                interpolation="none",
                cmap=grey_blue,
                vmin=0,
                vmax=1,
                extent=(left, pk - dpk, bottom, top),
                alpha=1,
            )
            ax.imshow(
                np.fliplr(Xe),
                interpolation="bicubic",
                cmap=grey_blue,
                vmin=0,
                vmax=1,
                extent=(pk - dpk - eps, pk + dpk, bottom, top),
                alpha=1,
            )
            ax.imshow(
                1 + Xe * 0,
                interpolation="none",
                cmap=grey_blue,
                vmin=0,
                vmax=1,
                extent=(pk + dpk - eps, right, bottom, top),
                alpha=1,
            )
            ax.text(
                pk - dtxt,
                i,
                " {:.2f} ".format(pk),
                color=pKa_color,
                fontsize=pKa_fontsize,
                horizontalalignment="right",
                zorder=30,
                path_effects=outline,
                weight="bold",
            )
        else:
            ax.imshow(
                1 + Xe * 0,
                interpolation="none",
                cmap=grey_red,
                vmin=0,
                vmax=1,
                extent=(left, pk - dpk, bottom, top),
                alpha=1,
            )
            ax.imshow(
                Xe,
                interpolation="bicubic",
                cmap=grey_red,
                vmin=0,
                vmax=1,
                extent=(pk - dpk - eps, pk + dpk, bottom, top),
                alpha=1,
            )
            ax.imshow(
                Xe * 0,
                interpolation="none",
                cmap=grey_red,
                vmin=0,
                vmax=1,
                extent=(pk + dpk - eps, right, bottom, top),
                alpha=1,
            )
            ax.text(
                pk + dtxt,
                i,
                " {:.2f} ".format(pk),
                color=pKa_color,
                fontsize=pKa_fontsize,
                horizontalalignment="left",
                zorder=30,
                path_effects=outline,
                weight="bold",
            )
        ax.add_line(
            Line2D([pk, pk], [bottom, top], linewidth=3, color="white", zorder=2)
        )

        # ax.add_line(Line2D([pk,pk], [bottom,top], linewidth=3, color='blue'))

    # Shaded vertical band at pH
    ax.axvline(x=pH - dpk, linewidth=2, color="black", alpha=0.2, linestyle="dashed")
    ax.axvline(x=pH + dpk, linewidth=2, color="black", alpha=0.2, linestyle="dashed")
    ax.axvline(x=pH, linewidth=3, color="black", alpha=0.5)
    ax.text(
        pH - dpk,
        ymax,
        " 90% protonated",
        rotation=90,
        horizontalalignment="right",
        verticalalignment="bottom",
        style="italic",
        path_effects=outline,
    )
    ax.text(
        pH + dpk,
        ymax,
        " 10% protonated",
        rotation=90,
        horizontalalignment="left",
        verticalalignment="bottom",
        style="italic",
        path_effects=outline,
    )

    ax.set_aspect("auto")
    plt.savefig(outname, dpi=300)
    plt.close(fig)


# The below is used for testing only
try:
    import aceprep
except ImportError:
    ACEPREP_EXISTS = False
else:
    ACEPREP_EXISTS = True


class _TestPreparation(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.home import home

        self.home = home(dataDir="test-systemprepare")

    def _compare_results(self, refpdb, refdf_f, pmol, df: pd.DataFrame):
        refdf = pd.read_csv(
            refdf_f, dtype=_table_dtypes, keep_default_na=False, na_values=[""]
        )
        refdf = refdf.fillna("")
        df = df.fillna("")
        try:
            pd.testing.assert_frame_equal(refdf, df)
        except AssertionError:
            if len(df) != len(refdf):
                raise AssertionError(
                    "Different number of residues were found in the DataFrames!"
                )
            for key in [
                "resname",
                "protonation",
                "resid",
                "insertion",
                "chain",
                "segid",
                "pKa",
                "buried",
            ]:
                refv = refdf[key].values
                newv = df[key].values
                diff = refv != newv
                if key in ("pKa", "buried"):
                    nans = refv == ""
                    refv = refv[~nans].astype(np.float32)
                    newv = newv[~nans].astype(np.float32)
                    diff = np.abs(refv - newv) > 1e-4
                if any(diff):
                    print(f"Found difference in field: {key}")
                    print(f"Reference values: {refv[diff]}")
                    print(f"New values:       {newv[diff]}")
            raise

        refmol = Molecule(refpdb)

        refmol.filter("not water")
        pmol.filter("not water")
        assert mol_equal(
            refmol, pmol, exceptFields=["serial"], fieldPrecision={"coords": 1e-3}
        )

    def test_systemPrepare(self):
        pdbids = ["3PTB", "1A25", "1U5U"]
        for pdb in pdbids:
            with self.subTest(pdbid=pdb):
                test_home = os.path.join(self.home, pdb)
                mol = Molecule(os.path.join(test_home, f"{pdb}.pdb"))
                pmol, df = systemPrepare(mol, return_details=True)
                self._compare_results(
                    os.path.join(test_home, f"{pdb}_prepared.pdb"),
                    os.path.join(test_home, f"{pdb}_prepared.csv"),
                    pmol,
                    df,
                )

    def test_systemprepare_ligand(self):
        test_home = os.path.join(self.home, "test-prepare-with-ligand")
        mol = Molecule(os.path.join(test_home, "5EK0_A.pdb"))
        pmol, df = systemPrepare(mol, return_details=True)
        self._compare_results(
            os.path.join(test_home, "5EK0_A_prepared.pdb"),
            os.path.join(test_home, "5EK0_A_prepared.csv"),
            pmol,
            df,
        )
        # Now remove the ligands and check again what the pka is
        mol.filter('not resname PX4 "5P2"')
        pmol, df = systemPrepare(mol, return_details=True)
        self._compare_results(
            os.path.join(test_home, "5EK0_A_prepared_nolig.pdb"),
            os.path.join(test_home, "5EK0_A_prepared_nolig.csv"),
            pmol,
            df,
        )

    def test_reprotonate(self):
        pmol, df = systemPrepare(Molecule("3PTB"), return_details=True)
        assert df.protonation[df.resid == 40].iloc[0] == "HIE"
        assert df.protonation[df.resid == 57].iloc[0] == "HIP"
        assert df.protonation[df.resid == 91].iloc[0] == "HID"

        pmol.mutateResidue("protein and resid 40", "HID")
        _, df2 = systemPrepare(pmol, titration=False, return_details=True)
        assert df2.protonation[df2.resid == 40].iloc[0] == "HID"
        assert df2.protonation[df2.resid == 57].iloc[0] == "HIP"
        assert df2.protonation[df2.resid == 91].iloc[0] == "HID"

        pmol, df = systemPrepare(
            Molecule("3PTB"),
            force_protonation=(
                ("protein and resid 40", "HID"),
                ("protein and resid 91", "HIE"),
            ),
            return_details=True,
        )
        assert df.protonation[df.resid == 40].iloc[0] == "HID"
        assert df.protonation[df.resid == 57].iloc[0] == "HIP"
        assert df.protonation[df.resid == 91].iloc[0] == "HIE"

    def test_auto_freezing(self):
        test_home = os.path.join(self.home, "test-auto-freezing")
        mol = Molecule(os.path.join(test_home, "2B5I.pdb"))

        pmol, df = systemPrepare(mol, return_details=True, hold_nonpeptidic_bonds=True)
        self._compare_results(
            os.path.join(test_home, "2B5I_prepared.pdb"),
            os.path.join(test_home, "2B5I_prepared.csv"),
            pmol,
            df,
        )

    def test_auto_freezing_and_force(self):
        test_home = os.path.join(self.home, "test-auto-freezing")
        mol = Molecule(os.path.join(test_home, "5DPX_A.pdb"))

        pmol, df = systemPrepare(
            mol,
            return_details=True,
            hold_nonpeptidic_bonds=True,
            force_protonation=[
                ("protein and resid 105 and chain A", "HIE"),
                ("protein and resid 107 and chain A", "HIE"),
                ("protein and resid 110 and chain A", "HIE"),
                ("protein and resid 181 and chain A", "HIE"),
                ("protein and resid 246 and chain A", "HIE"),
            ],
        )
        self._compare_results(
            os.path.join(test_home, "5DPX_A_prepared.pdb"),
            os.path.join(test_home, "5DPX_A_prepared.csv"),
            pmol,
            df,
        )

    def test_nonstandard_residues(self):
        test_home = os.path.join(self.home, "test-nonstandard-residues")
        mol = Molecule(os.path.join(test_home, "1A4W.pdb"))

        pmol, df = systemPrepare(mol, return_details=True, hold_nonpeptidic_bonds=True)

        self._compare_results(
            os.path.join(test_home, "1A4W_prepared.pdb"),
            os.path.join(test_home, "1A4W_prepared.csv"),
            pmol,
            df,
        )

    @unittest.skipUnless(ACEPREP_EXISTS, "Can only run with aceprep installed")
    def test_unknown_nonstandard_residues(self):
        test_home = os.path.join(self.home, "test-nonstandard-residues")
        mol = Molecule(os.path.join(test_home, "1A4W.pdb"))

        pmol, df = systemPrepare(
            mol, return_details=True, hold_nonpeptidic_bonds=True, _molkit_ff=False
        )

        self._compare_results(
            os.path.join(test_home, "1A4W_prepared.pdb"),
            os.path.join(test_home, "1A4W_prepared.csv"),
            pmol,
            df,
        )

    def test_rna_protein_complex(self):
        test_home = os.path.join(self.home, "test-rna-protein-complex")
        mol = Molecule(os.path.join(test_home, "3WBM.pdb"))

        pmol, df = systemPrepare(mol, return_details=True)

        self._compare_results(
            os.path.join(test_home, "3WBM_prepared.pdb"),
            os.path.join(test_home, "3WBM_prepared.csv"),
            pmol,
            df,
        )

    def test_dna(self):
        test_home = os.path.join(self.home, "test-dna")
        mol = Molecule(os.path.join(test_home, "1BNA.pdb"))

        pmol, df = systemPrepare(mol, return_details=True)

        self._compare_results(
            os.path.join(test_home, "1BNA_prepared.pdb"),
            os.path.join(test_home, "1BNA_prepared.csv"),
            pmol,
            df,
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)
    import doctest

    doctest.testmod()



##
# File:  PdbxWriter.py
# Date:  2011-10-09 Jdw  Adapted from PdbxParser.py
#
# Updates:
#    5-Apr-2011 jdw  Using the double quote format preference
#   23-Oct-2012 jdw  update path details and reorganize.
#
###
"""
Classes for writing data and dictionary containers in PDBx/mmCIF format.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import sys
from moleculekit.pdbx.reader.PdbxContainers import DefinitionContainer, DataContainer


class PdbxError(Exception):
    """Class for catch general errors"""

    pass


class PdbxWriter(object):
    """Write PDBx data files or dictionaries using the input container
    or container list.
    """

    def __init__(self, ofh=sys.stdout):
        self.__ofh = ofh
        self.__containerList = []
        self.__MAXIMUM_LINE_LENGTH = 2048
        self.__SPACING = 2
        self.__INDENT_DEFINITION = 3
        self.__indentSpace = " " * self.__INDENT_DEFINITION
        self.__doDefinitionIndent = False
        # Maximum number of rows checked for value length and format
        self.__rowPartition = None

    def setRowPartition(self, numRows):
        """Maximum number of rows checked for value length and format"""
        self.__rowPartition = numRows

    def write(self, containerList):
        self.__containerList = containerList
        for container in self.__containerList:
            self.writeContainer(container)

    def writeContainer(self, container):
        indS = " " * self.__INDENT_DEFINITION
        if isinstance(container, DefinitionContainer):
            self.__write("save_%s\n" % container.getName())
            self.__doDefinitionIndent = True
            self.__write(indS + "#\n")
        elif isinstance(container, DataContainer):
            if container.getGlobal():
                self.__write("global_\n")
                self.__doDefinitionIndent = False
                self.__write("\n")
            else:
                self.__write("data_%s\n" % container.getName())
                self.__doDefinitionIndent = False
                self.__write("#\n")

        for nm in container.getObjNameList():
            obj = container.getObj(nm)
            objL = obj.getRowList()

            # Skip empty objects
            if len(objL) == 0:
                continue

            # Item - value formattting
            elif len(objL) == 1:
                self.__writeItemValueFormat(obj)

            # Table formatting -
            elif len(objL) > 1 and len(obj.getAttributeList()) > 0:
                self.__writeTableFormat(obj)
            else:
                raise PdbxError()

            if self.__doDefinitionIndent:
                self.__write(indS + "#")
            else:
                self.__write("#")

        # Add a trailing saveframe reserved word
        if isinstance(container, DefinitionContainer):
            self.__write("\nsave_\n")
        self.__write("#\n")

    def __write(self, st):
        self.__ofh.write(st)

    def __writeItemValueFormat(self, myCategory):

        # Compute the maximum item name length within this category -
        attributeNameLengthMax = 0
        for attributeName in myCategory.getAttributeList():
            attributeNameLengthMax = max(attributeNameLengthMax, len(attributeName))
        itemNameLengthMax = (
            self.__SPACING + len(myCategory.getName()) + attributeNameLengthMax + 2
        )
        #
        lineList = []
        lineList.append("#\n")
        for attributeName, iPos in myCategory.getAttributeListWithOrder():
            if self.__doDefinitionIndent:
                #        - add indent --
                lineList.append(self.__indentSpace)

            itemName = "_%s.%s" % (myCategory.getName(), attributeName)
            lineList.append(itemName.ljust(itemNameLengthMax))

            lineList.append(myCategory.getValueFormatted(attributeName, 0))
            lineList.append("\n")

        self.__write("".join(lineList))

    def __writeTableFormat(self, myCategory):

        # Write the declaration of the loop_
        #
        lineList = []
        lineList.append("#\n")
        if self.__doDefinitionIndent:
            lineList.append(self.__indentSpace)
        lineList.append("loop_")
        for attributeName in myCategory.getAttributeList():
            lineList.append("\n")
            if self.__doDefinitionIndent:
                lineList.append(self.__indentSpace)
            itemName = "_%s.%s" % (myCategory.getName(), attributeName)
            lineList.append(itemName)
        self.__write("".join(lineList))

        #
        # Write the data in tabular format -
        #
        # print myCategory.getName()
        # print myCategory.getAttributeList()

        #    For speed make the following evaluation on a portion of the table
        if self.__rowPartition is not None:
            numSteps = max(1, myCategory.getRowCount() / self.__rowPartition)
        else:
            numSteps = 1

        formatTypeList, dataTypeList = myCategory.getFormatTypeList(steps=numSteps)
        maxLengthList = myCategory.getAttributeValueMaxLengthList(steps=numSteps)
        spacing = " " * self.__SPACING
        #

        # print formatTypeList
        # print dataTypeList
        # print maxLengthList
        #
        for iRow in range(myCategory.getRowCount()):
            lineList = []
            lineList.append("\n")
            if self.__doDefinitionIndent:
                lineList.append(self.__indentSpace + "  ")

            for iAt in range(myCategory.getAttributeCount()):
                formatType = formatTypeList[iAt]
                maxLength = maxLengthList[iAt]

                if formatType == "FT_UNQUOTED_STRING" or formatType == "FT_NULL_VALUE":
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val.ljust(maxLength))

                elif formatType == "FT_NUMBER":
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val.rjust(maxLength))

                elif formatType == "FT_QUOTED_STRING":
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)

                    lineList.append(val.ljust(maxLength + 2))

                elif formatType == "FT_MULTI_LINE_STRING":
                    val = myCategory.getValueFormattedByIndex(iAt, iRow)
                    lineList.append(val)

                lineList.append(spacing)

            self.__write("".join(lineList))
        self.__write("\n")



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import os

_config = {
    "viewer": "VMD",
    "configfile": os.getenv("HTMD_CONFIG") if os.getenv("HTMD_CONFIG") else None,
    "ncpus": 1,
}


def config(
    viewer=_config["viewer"], configfile=_config["configfile"], ncpus=_config["ncpus"]
):
    """
    Function to change HTMD configuration variables.

    Parameters
    ----------
    viewer : str
        Defines the backend viewer for molecular visualization
    ncpus : int
        Defines the number of cpus available for several HTMD operations
    configfile : str
        Defines the HTMD configuration file that is called at the beginning of importing
    lsf : str
        Defines a YAML file that can contain default profile configurations for an LsfQueue
    slurm : str
        Defines a YAML file that can contain default profile configurations for an SlurmQueue
    """
    _config["viewer"] = viewer
    _config["configfile"] = configfile
    _config["ncpus"] = ncpus



from collections import namedtuple
import logging

logger = logging.getLogger(__file__)

# from periodictable import elements  # the python package called periodictable, not this file
# for el in elements:
#     if el.name == 'neutron':
#         continue
#     # if mendeleev.__dict__[el.symbol].vdw_radius_bondi is not None:
#     #     vdw_radius = mendeleev.__dict__[el.symbol].vdw_radius_bondi
#     # else:
#     vdw_radius = mendeleev.__dict__[el.symbol].vdw_radius
#     if vdw_radius is not None:
#         vdw_radius /= 100  # Convert picometer to A
#     print("'{}': _Element(symbol='{}', name='{}', number={}, mass={}, vdw_radius={}),".format(el, el, el.name, el.number, el.mass, vdw_radius))


_Element = namedtuple("Element", ["symbol", "name", "number", "mass", "vdw_radius"])
periodictable = {
    "H": _Element(symbol="H", name="hydrogen", number=1, mass=1.00794, vdw_radius=1.1),
    "He": _Element(symbol="He", name="helium", number=2, mass=4.002602, vdw_radius=1.4),
    "Li": _Element(symbol="Li", name="lithium", number=3, mass=6.941, vdw_radius=1.82),
    "Be": _Element(
        symbol="Be", name="beryllium", number=4, mass=9.012182, vdw_radius=1.53
    ),
    "B": _Element(symbol="B", name="boron", number=5, mass=10.811, vdw_radius=1.92),
    "C": _Element(symbol="C", name="carbon", number=6, mass=12.0107, vdw_radius=1.7),
    "N": _Element(symbol="N", name="nitrogen", number=7, mass=14.0067, vdw_radius=1.55),
    "O": _Element(symbol="O", name="oxygen", number=8, mass=15.9994, vdw_radius=1.52),
    "F": _Element(
        symbol="F", name="fluorine", number=9, mass=18.9984032, vdw_radius=1.47
    ),
    "Ne": _Element(symbol="Ne", name="neon", number=10, mass=20.1797, vdw_radius=1.54),
    "Na": _Element(
        symbol="Na", name="sodium", number=11, mass=22.98977, vdw_radius=2.27
    ),
    "Mg": _Element(
        symbol="Mg", name="magnesium", number=12, mass=24.305, vdw_radius=1.73
    ),
    "Al": _Element(
        symbol="Al", name="aluminum", number=13, mass=26.981538, vdw_radius=1.84
    ),
    "Si": _Element(
        symbol="Si", name="silicon", number=14, mass=28.0855, vdw_radius=2.1
    ),
    "P": _Element(
        symbol="P", name="phosphorus", number=15, mass=30.973761, vdw_radius=1.8
    ),
    "S": _Element(symbol="S", name="sulfur", number=16, mass=32.065, vdw_radius=1.8),
    "Cl": _Element(
        symbol="Cl", name="chlorine", number=17, mass=35.453, vdw_radius=1.75
    ),
    "Ar": _Element(symbol="Ar", name="argon", number=18, mass=39.948, vdw_radius=1.88),
    "K": _Element(
        symbol="K", name="potassium", number=19, mass=39.0983, vdw_radius=2.75
    ),
    "Ca": _Element(
        symbol="Ca", name="calcium", number=20, mass=40.078, vdw_radius=2.31
    ),
    "Sc": _Element(
        symbol="Sc", name="scandium", number=21, mass=44.95591, vdw_radius=2.15
    ),
    "Ti": _Element(
        symbol="Ti", name="titanium", number=22, mass=47.867, vdw_radius=2.11
    ),
    "V": _Element(
        symbol="V", name="vanadium", number=23, mass=50.9415, vdw_radius=2.07
    ),
    "Cr": _Element(
        symbol="Cr", name="chromium", number=24, mass=51.9961, vdw_radius=2.06
    ),
    "Mn": _Element(
        symbol="Mn", name="manganese", number=25, mass=54.938049, vdw_radius=2.05
    ),
    "Fe": _Element(symbol="Fe", name="iron", number=26, mass=55.845, vdw_radius=2.04),
    "Co": _Element(symbol="Co", name="cobalt", number=27, mass=58.9332, vdw_radius=2.0),
    "Ni": _Element(
        symbol="Ni", name="nickel", number=28, mass=58.6934, vdw_radius=1.97
    ),
    "Cu": _Element(symbol="Cu", name="copper", number=29, mass=63.546, vdw_radius=1.96),
    "Zn": _Element(symbol="Zn", name="zinc", number=30, mass=65.409, vdw_radius=2.01),
    "Ga": _Element(
        symbol="Ga", name="gallium", number=31, mass=69.723, vdw_radius=1.87
    ),
    "Ge": _Element(
        symbol="Ge", name="germanium", number=32, mass=72.64, vdw_radius=2.11
    ),
    "As": _Element(
        symbol="As", name="arsenic", number=33, mass=74.9216, vdw_radius=1.85
    ),
    "Se": _Element(symbol="Se", name="selenium", number=34, mass=78.96, vdw_radius=1.9),
    "Br": _Element(
        symbol="Br", name="bromine", number=35, mass=79.904, vdw_radius=1.85
    ),
    "Kr": _Element(
        symbol="Kr", name="krypton", number=36, mass=83.798, vdw_radius=2.02
    ),
    "Rb": _Element(
        symbol="Rb", name="rubidium", number=37, mass=85.4678, vdw_radius=3.03
    ),
    "Sr": _Element(
        symbol="Sr", name="strontium", number=38, mass=87.62, vdw_radius=2.49
    ),
    "Y": _Element(
        symbol="Y", name="yttrium", number=39, mass=88.90585, vdw_radius=2.32
    ),
    "Zr": _Element(
        symbol="Zr", name="zirconium", number=40, mass=91.224, vdw_radius=2.23
    ),
    "Nb": _Element(
        symbol="Nb", name="niobium", number=41, mass=92.90638, vdw_radius=2.18
    ),
    "Mo": _Element(
        symbol="Mo", name="molybdenum", number=42, mass=95.94, vdw_radius=2.17
    ),
    "Tc": _Element(symbol="Tc", name="technetium", number=43, mass=98, vdw_radius=2.16),
    "Ru": _Element(
        symbol="Ru", name="ruthenium", number=44, mass=101.07, vdw_radius=2.13
    ),
    "Rh": _Element(
        symbol="Rh", name="rhodium", number=45, mass=102.9055, vdw_radius=2.1
    ),
    "Pd": _Element(
        symbol="Pd", name="palladium", number=46, mass=106.42, vdw_radius=2.1
    ),
    "Ag": _Element(
        symbol="Ag", name="silver", number=47, mass=107.8682, vdw_radius=2.11
    ),
    "Cd": _Element(
        symbol="Cd", name="cadmium", number=48, mass=112.411, vdw_radius=2.18
    ),
    "In": _Element(
        symbol="In", name="indium", number=49, mass=114.818, vdw_radius=1.93
    ),
    "Sn": _Element(symbol="Sn", name="tin", number=50, mass=118.71, vdw_radius=2.17),
    "Sb": _Element(
        symbol="Sb", name="antimony", number=51, mass=121.76, vdw_radius=2.06
    ),
    "Te": _Element(
        symbol="Te", name="tellurium", number=52, mass=127.6, vdw_radius=2.06
    ),
    "I": _Element(
        symbol="I", name="iodine", number=53, mass=126.90447, vdw_radius=1.98
    ),
    "Xe": _Element(symbol="Xe", name="xenon", number=54, mass=131.293, vdw_radius=2.16),
    "Cs": _Element(
        symbol="Cs", name="cesium", number=55, mass=132.90545, vdw_radius=3.43
    ),
    "Ba": _Element(
        symbol="Ba", name="barium", number=56, mass=137.327, vdw_radius=2.68
    ),
    "La": _Element(
        symbol="La", name="lanthanum", number=57, mass=138.9055, vdw_radius=2.43
    ),
    "Ce": _Element(
        symbol="Ce", name="cerium", number=58, mass=140.116, vdw_radius=2.42
    ),
    "Pr": _Element(
        symbol="Pr", name="praseodymium", number=59, mass=140.90765, vdw_radius=2.4
    ),
    "Nd": _Element(
        symbol="Nd", name="neodymium", number=60, mass=144.24, vdw_radius=2.39
    ),
    "Pm": _Element(
        symbol="Pm", name="promethium", number=61, mass=145, vdw_radius=2.38
    ),
    "Sm": _Element(
        symbol="Sm", name="samarium", number=62, mass=150.36, vdw_radius=2.36
    ),
    "Eu": _Element(
        symbol="Eu", name="europium", number=63, mass=151.964, vdw_radius=2.35
    ),
    "Gd": _Element(
        symbol="Gd", name="gadolinium", number=64, mass=157.25, vdw_radius=2.34
    ),
    "Tb": _Element(
        symbol="Tb", name="terbium", number=65, mass=158.92534, vdw_radius=2.33
    ),
    "Dy": _Element(
        symbol="Dy", name="dysprosium", number=66, mass=162.5, vdw_radius=2.31
    ),
    "Ho": _Element(
        symbol="Ho", name="holmium", number=67, mass=164.93032, vdw_radius=2.3
    ),
    "Er": _Element(
        symbol="Er", name="erbium", number=68, mass=167.259, vdw_radius=2.29
    ),
    "Tm": _Element(
        symbol="Tm", name="thulium", number=69, mass=168.93421, vdw_radius=2.27
    ),
    "Yb": _Element(
        symbol="Yb", name="ytterbium", number=70, mass=173.04, vdw_radius=2.26
    ),
    "Lu": _Element(
        symbol="Lu", name="lutetium", number=71, mass=174.967, vdw_radius=2.24
    ),
    "Hf": _Element(
        symbol="Hf", name="hafnium", number=72, mass=178.49, vdw_radius=2.23
    ),
    "Ta": _Element(
        symbol="Ta", name="tantalum", number=73, mass=180.9479, vdw_radius=2.22
    ),
    "W": _Element(symbol="W", name="tungsten", number=74, mass=183.84, vdw_radius=2.18),
    "Re": _Element(
        symbol="Re", name="rhenium", number=75, mass=186.207, vdw_radius=2.16
    ),
    "Os": _Element(symbol="Os", name="osmium", number=76, mass=190.23, vdw_radius=2.16),
    "Ir": _Element(
        symbol="Ir", name="iridium", number=77, mass=192.217, vdw_radius=2.13
    ),
    "Pt": _Element(
        symbol="Pt", name="platinum", number=78, mass=195.078, vdw_radius=2.13
    ),
    "Au": _Element(
        symbol="Au", name="gold", number=79, mass=196.96655, vdw_radius=2.14
    ),
    "Hg": _Element(
        symbol="Hg", name="mercury", number=80, mass=200.59, vdw_radius=2.23
    ),
    "Tl": _Element(
        symbol="Tl", name="thallium", number=81, mass=204.3833, vdw_radius=1.96
    ),
    "Pb": _Element(symbol="Pb", name="lead", number=82, mass=207.2, vdw_radius=2.02),
    "Bi": _Element(
        symbol="Bi", name="bismuth", number=83, mass=208.98038, vdw_radius=2.07
    ),
    "Po": _Element(symbol="Po", name="polonium", number=84, mass=209, vdw_radius=1.97),
    "At": _Element(symbol="At", name="astatine", number=85, mass=210, vdw_radius=2.02),
    "Rn": _Element(symbol="Rn", name="radon", number=86, mass=222, vdw_radius=2.2),
    "Fr": _Element(symbol="Fr", name="francium", number=87, mass=223, vdw_radius=3.48),
    "Ra": _Element(symbol="Ra", name="radium", number=88, mass=226, vdw_radius=2.83),
    "Ac": _Element(symbol="Ac", name="actinium", number=89, mass=227, vdw_radius=2.47),
    "Th": _Element(
        symbol="Th", name="thorium", number=90, mass=232.0381, vdw_radius=2.45
    ),
    "Pa": _Element(
        symbol="Pa", name="protactinium", number=91, mass=231.03588, vdw_radius=2.43
    ),
    "U": _Element(
        symbol="U", name="uranium", number=92, mass=238.02891, vdw_radius=2.41
    ),
    "Np": _Element(symbol="Np", name="neptunium", number=93, mass=237, vdw_radius=2.39),
    "Pu": _Element(symbol="Pu", name="plutonium", number=94, mass=244, vdw_radius=2.43),
    "Am": _Element(symbol="Am", name="americium", number=95, mass=243, vdw_radius=2.44),
    "Cm": _Element(symbol="Cm", name="curium", number=96, mass=247, vdw_radius=2.45),
    "Bk": _Element(symbol="Bk", name="berkelium", number=97, mass=247, vdw_radius=2.44),
    "Cf": _Element(
        symbol="Cf", name="californium", number=98, mass=251, vdw_radius=2.45
    ),
    "Es": _Element(
        symbol="Es", name="einsteinium", number=99, mass=252, vdw_radius=2.45
    ),
    "Fm": _Element(symbol="Fm", name="fermium", number=100, mass=257, vdw_radius=2.45),
    "Md": _Element(
        symbol="Md", name="mendelevium", number=101, mass=258, vdw_radius=2.46
    ),
    "No": _Element(symbol="No", name="nobelium", number=102, mass=259, vdw_radius=2.46),
    "Lr": _Element(
        symbol="Lr", name="lawrencium", number=103, mass=262, vdw_radius=2.46
    ),
    "Rf": _Element(
        symbol="Rf", name="rutherfordium", number=104, mass=261, vdw_radius=None
    ),
    "Db": _Element(symbol="Db", name="dubnium", number=105, mass=262, vdw_radius=None),
    "Sg": _Element(
        symbol="Sg", name="seaborgium", number=106, mass=266, vdw_radius=None
    ),
    "Bh": _Element(symbol="Bh", name="bohrium", number=107, mass=264, vdw_radius=None),
    "Hs": _Element(symbol="Hs", name="hassium", number=108, mass=277, vdw_radius=None),
    "Mt": _Element(
        symbol="Mt", name="meitnerium", number=109, mass=268, vdw_radius=None
    ),
    "Ds": _Element(
        symbol="Ds", name="darmstadtium", number=110, mass=281, vdw_radius=None
    ),
    "Rg": _Element(
        symbol="Rg", name="roentgenium", number=111, mass=272, vdw_radius=None
    ),
    "Cn": _Element(
        symbol="Cn", name="copernicium", number=112, mass=285, vdw_radius=None
    ),
    "Nh": _Element(symbol="Nh", name="nihonium", number=113, mass=286, vdw_radius=None),
    "Fl": _Element(
        symbol="Fl", name="flerovium", number=114, mass=289, vdw_radius=None
    ),
    "Mc": _Element(
        symbol="Mc", name="moscovium", number=115, mass=289, vdw_radius=None
    ),
    "Lv": _Element(
        symbol="Lv", name="livermorium", number=116, mass=293, vdw_radius=None
    ),
    "Ts": _Element(
        symbol="Ts", name="tennessine", number=117, mass=294, vdw_radius=None
    ),
    "Og": _Element(
        symbol="Og", name="oganesson", number=118, mass=294, vdw_radius=None
    ),
}

# Add indexes as well by atomic number
periodictable_by_number = {
    periodictable[el].number: val for el, val in periodictable.items()
}

import numpy as np

# This of course fails for exotic elements like Bk-Cm Db-Lr Mc-Fl Og-Ts which have similar masses
_all_elements = np.array([el for el in periodictable])
_all_masses = np.array([periodictable[el].mass for el in periodictable])


def elements_from_masses(masses):
    from moleculekit.util import ensurelist
    from scipy.spatial.distance import cdist

    masses = np.array(ensurelist(masses))
    if np.any(masses > 140):
        logger.warning(
            "Guessing element for atoms with mass > 140. This can lead to inaccurate element guesses."
        )
    elements = list(
        _all_elements[
            np.argmin(cdist(masses[:, None], np.array(_all_masses)[:, None]), axis=1)
        ]
    )
    if len(elements) == 1:
        return elements[0]
    return elements


import unittest


class _TestPeriodicTable(unittest.TestCase):
    def test_elements_from_masses(self):
        # Only test lower masses. The high ones are not very exact
        masses_to_test = _all_masses[_all_masses < 140]
        elements_to_test = _all_elements[_all_masses < 140]
        assert np.array_equal(elements_to_test, elements_from_masses(masses_to_test))
        assert np.array_equal(
            elements_to_test, elements_from_masses(masses_to_test + 0.05)
        )


if __name__ == "__main__":
    unittest.main(verbosity=2)



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import tempfile
import re
import numpy as np
import subprocess

try:
    import pandas
except ImportError:
    print("Pandas is not installed. Please install it with `conda install pandas`")


def getSequenceProfile(sequence, hhblits, hhblitsdb, ncpu=6, niter=4):
    """Calculates the sequence profile of a protein sequence using HHBlits

    File description and unit conversions taken from section 6 https://hpc.nih.gov/apps/hhsuite-userguide.pdf

    Parameters
    ----------
    sequence : str
        A string encoding the one letter sequence of the protein
    hhblits : str
        The path to the hhblits executable
    hhblitsdb : str
        The path to the hhblits database that we want to search against. Should include the database name prefix, not just the folder.
    ncpu : int
        Number of CPUs to use for the search
    niter : int
        The number of hhblits iterations. The higher the value the more remote homologues it will find

    Returns
    -------
    df : pandas.DataFrame
        A pandas dataframe containing all the information read from the file
    pssm : np.ndarray
        A Nx20 numpy array where N the number of residues of the protein. Contains the transition probabilities to all 20 residues.

    Examples
    --------
    >>> hhb = '~/hhsuite-2.0.16-linux-x86_64/bin/hhblits'
    >>> hhbdb = '~/hhsuite-2.0.16-linux-x86_64/databases/uniprot20_2016_02/uniprot20_2016_02'
    >>> seq = 'MKVIFLKDVKGMGKKGEIKNVADGYANNFLFKQGLAIEA'
    >>> df, prof = getSequenceProfile(seq, hhb, hhbdb)
    """
    regex = re.compile("^\w\s\d+")
    with tempfile.TemporaryDirectory() as tmpdir:
        with open(tmpdir + "/input.fasta", "w") as fp:
            fp.write(f">\n{sequence}\n")

        try:
            subprocess.check_output(
                "{} -i {}/input.fasta -d {} -cpu {} -M first -n {} -ohhm {}/output.hhm -v 0".format(
                    hhblits, tmpdir, hhblitsdb, ncpu, niter, tmpdir
                ),
                stderr=subprocess.STDOUT,
                shell=True,
            )
        except subprocess.CalledProcessError as err:
            print(
                "HHBlits errored out with code {} and ouput: \n--------------- ERROR LOG ---------------\n{}\n--------------- END OF LOG ---------------\n".format(
                    err.returncode, err.output.decode("utf-8")
                )
            )

        data = []
        seq = []
        with open(tmpdir + "/output.hhm", "r") as fp:
            starting = 0
            lines = fp.readlines()
            for i in range(len(lines)):
                if lines[i].startswith("NULL"):
                    pieces = lines[i].split()
                    seq.append([pieces[0]])
                    data.append(
                        [2 ** (-int(x) / 1000) if x != "*" else 0 for x in pieces[1:21]]
                        + [0] * 10
                    )
                if lines[i].startswith("HMM    A	C	D"):
                    col_desc = lines[i].split()[1:] + lines[i + 1].split()
                    starting = 1
                if starting > 0:
                    starting += 1
                if starting >= 4 and regex.match(lines[i]):
                    pieces = lines[i].split()
                    seq.append([pieces[0]])
                    d = [2 ** (-int(x) / 1000) if x != "*" else 0 for x in pieces[2:22]]
                    pieces = lines[i + 1].split()
                    d += [2 ** (-int(x) / 1000) if x != "*" else 0 for x in pieces[:7]]
                    d += [0.001 * int(x) for x in pieces[7:10]]
                    data.append(d)

        df = pandas.DataFrame(
            np.hstack((np.vstack(seq), np.vstack(data))), columns=["seq"] + col_desc
        )
        return df, df.values[1:, 1:21].astype(float)


if __name__ == "__main__":
    pass



import os
import unittest
from tempfile import NamedTemporaryFile
from moleculekit.home import home
from moleculekit.smallmol.smallmol import SmallMol
from moleculekit.smallmol.smallmollib import SmallMolLib
import rdkit
from pandas import core


SDF_N_MOLS = 100

SDF_IDS_DELETE = [1, 10, 15, 16]
SDF_MOLNAME_DELETE = ["ZINC02583363", "ZINC86860147", "ZINC04342657", "ZINC02023420"]
SDF_FIELDS = ["ligname", "_mol"]
SDF_LOC_0_99 = "ZINC02141008"


class _TestSmallMol(unittest.TestCase):
    def setUp(self):
        self.dataDir = home("test-smallmol")

    def test_loaSdffile(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        n_mols = lib.numMols
        self.assertEqual(
            n_mols,
            SDF_N_MOLS,
            f"Molecules not correctly loaded. Expected: {SDF_N_MOLS}; Now: {n_mols}",
        )

        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf.gz")
        lib = SmallMolLib(sdffile)
        n_mols = lib.numMols
        self.assertEqual(
            n_mols,
            SDF_N_MOLS,
            f"Molecules not correctly loaded. Expected: {SDF_N_MOLS}; Now: {n_mols}",
        )

    def test_writeSdf(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)

        sdfname = NamedTemporaryFile().name + ".sdf"
        lib.writeSdf(sdfname)

        sdf_exists = os.path.isfile(sdfname)

        self.assertTrue(sdf_exists, msg="The sdf written was not found")

        sdf = SmallMolLib(sdfname)

        self.assertIsInstance(
            sdf,
            SmallMolLib,
            msg="The sdf written was not correctly loaded. Probably the previous"
            "writing went wrong",
        )

    def test_appendSmallMolLib(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        lib2 = SmallMolLib(sdffile)

        lib.appendSmallLib(lib2)

        n_mol2_merged = lib.numMols

        self.assertEqual(
            n_mol2_merged,
            SDF_N_MOLS * 2,
            msg="The number of molecules in the SmallMolLib is not as expected."
            "The two sdf were not correctly merged. ",
        )

    def test_appendSmallMol(self):
        mol2file = os.path.join(self.dataDir, "benzamidine.mol2")
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")

        lib = SmallMolLib(sdffile)
        sm = SmallMol(mol2file)
        lib.appendSmallMol(sm)

        n_mol2_append = lib.numMols

        self.assertEqual(
            n_mol2_append,
            SDF_N_MOLS + 1,
            msg="The number of molecules in the SmallMolLib is not as expected."
            "The mol2 were not correctly append. ",
        )

    def test_removeMols(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")

        lib = SmallMolLib(sdffile)

        mols_ids = SDF_IDS_DELETE
        ref_mols_name = SDF_MOLNAME_DELETE

        mols_name = [s.ligname for s in lib.getMols(mols_ids)]

        self.assertListEqual(
            mols_name,
            ref_mols_name,
            msg="The molecules at the given indexes do not match with the" "expected",
        )
        lib.removeMols(mols_ids)

        mols_name_now = [s.ligname for s in lib.getMols(mols_ids)]

        self.assertFalse(
            mols_name_now == mols_name,
            msg="The molecules seem to not be deleted correctly",
        )

    def test_convertToDataFrame(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")

        lib = SmallMolLib(sdffile)

        df = lib.toDataFrame()

        self.assertIsInstance(
            df,
            core.frame.DataFrame,
            msg="The SmallMolLib object was not correctly converted into pandas"
            "DataFrame",
        )

        cols = df.columns.tolist()
        ref_cols = SDF_FIELDS

        self.assertEqual(
            cols,
            ref_cols,
            msg="The fields in the SmallMolLib object was not the expected one",
        )

        ligname_99 = df.iloc[99][0]
        ref_ligname = SDF_LOC_0_99

        self.assertEqual(
            ligname_99, ref_ligname, msg="The ligand name found is not the expected one"
        )

    def test_writeSmiles(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")

        lib = SmallMolLib(sdffile)
        tmpfile = NamedTemporaryFile().name + ".smi"
        lib.writeSmiles(tmpfile)

        with open(tmpfile, "r") as f:
            filelines = f.readlines()[1:]

        with open(os.path.join(self.dataDir, "fda_drugs_light.smi"), "r") as f:
            reflines = f.readlines()[1:]

        self.assertEqual(filelines, reflines)

    def test_readSmiles(self):
        smifile = os.path.join(self.dataDir, "fda_drugs_light.smi")
        lib = SmallMolLib(smifile)
        assert len(lib) == 100

        smifile = os.path.join(self.dataDir, "fda_drugs_light.smi.gz")
        lib = SmallMolLib(smifile)
        assert len(lib) == 100


if __name__ == "__main__":
    unittest.main(verbosity=2)



import os
import unittest
from moleculekit.home import home
from moleculekit.smallmol.smallmollib import SmallMolLib
from moleculekit.smallmol.tools.clustering import cluster, getMaximumCommonSubstructure
import numpy as np
import rdkit


CMS_N_ATOMS = 2
CMS_AIDX_MOL0 = [0, 5]

# fmt: off
# clusters
MACCS_N_CLUSTER = 75
MACCS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 3, 1, 2, 1, 1, 1, 1, 1, 2, 3, 2, 2, 2, 1, 1, 1, 1, 2, 3, 
                            1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 2, 2, 1, 1, 
                            1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]

PATHFINGERPRINTS_N_CLUSTER = 82
PATHFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 1, 1, 
                                       1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 
                                       1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]

ATOMSFINGERPRINTS_N_CLUSTER = 79
ATOMSFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 1, 1, 1, 
                                        1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 
                                        1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]

TORSIONSFINGERPRINTS_N_CLUSTER = 79
TORSIONSFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 1, 
                                           1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 2, 1, 1, 2, 
                                           1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1]


CIRCULARFINGERPRINTS_N_CLUSTER = 80
CIRCULARFINGERPRINTS_POPULATION_CLUSTER = [2, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 3, 1, 2, 2, 1, 
                                           1, 1, 1, 2, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 2, 1, 1, 2, 
                                           1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1]

MCS_N_CLUSTER = 77
MCS_POPULATION_CLUSTER = [2, 1, 2, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2, 3, 1, 2, 2, 2, 1, 1, 1, 2, 2, 1, 1, 2, 1, 
                          1, 1, 1, 1, 1, 2, 1, 3, 2, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 
                          2, 1, 1, 1, 1, 1, 1, 1]
# fmt: on


class _TestSmallMol(unittest.TestCase):
    def setUp(self):
        self.dataDir = home("test-smallmol")

    def test_getCommonStructure(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        rd_mols = lib._mols

        cms, cms_idxs, _ = getMaximumCommonSubstructure(rd_mols, returnAtomIdxs=True)

        self.assertIsInstance(
            cms, rdkit.Chem.rdchem.Mol, msg="The object is not a rdkit Molecule object"
        )

        ref_cms_natoms = CMS_N_ATOMS
        ref_atomidx_mol0 = CMS_AIDX_MOL0
        cms_atomidx_mol0 = cms_idxs[0]
        cms_natoms = len(cms_atomidx_mol0)

        self.assertEqual(
            cms_atomidx_mol0,
            ref_atomidx_mol0,
            msg="The atomidx of the CMS for the mol0 are not the expected ones",
        )
        self.assertEqual(
            cms_natoms,
            ref_cms_natoms,
            msg="The number of atom in the CMS for the mol0 are not the expected ones",
        )

    _methods = [
        "maccs",
        "pathFingerprints",
        "atomsFingerprints",
        "torsionsFingerprints",
        "circularFingerprints",
        "shape",
        "mcs",
    ]

    def test_cluster_maccs(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        _, det = cluster(mols, "maccs", returnDetails=True)

        ref_ncluster = MACCS_N_CLUSTER
        ref_populations = MACCS_POPULATION_CLUSTER

        ncluster = det["numClusters"]
        population = det["populations"].tolist()

        self.assertEqual(
            ncluster,
            ref_ncluster,
            msg="The number of cluster identified are not as expected",
        )
        self.assertEqual(
            population,
            ref_populations,
            msg="The population fo the cluster are not the expected one",
        )

    def test_cluster_pathFingerprints(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        _, det = cluster(mols, "pathFingerprints", returnDetails=True)

        ref_ncluster = PATHFINGERPRINTS_N_CLUSTER
        ref_populations = PATHFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det["numClusters"]
        population = det["populations"].tolist()

        self.assertEqual(
            ncluster,
            ref_ncluster,
            msg="The number of cluster identified are not as expected",
        )
        self.assertEqual(
            population,
            ref_populations,
            msg="The population fo the cluster are not the expected one",
        )

    def test_cluster_atomsFingerprints(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        _, det = cluster(mols, "atomsFingerprints", returnDetails=True)

        ref_ncluster = ATOMSFINGERPRINTS_N_CLUSTER
        ref_populations = ATOMSFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det["numClusters"]
        population = det["populations"].tolist()

        self.assertEqual(
            ncluster,
            ref_ncluster,
            msg="The number of cluster identified are not as expected",
        )
        self.assertEqual(
            population,
            ref_populations,
            msg="The population fo the cluster are not the expected one",
        )

    def test_cluster_torsionsFingerprints(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        _, det = cluster(mols, "torsionsFingerprints", returnDetails=True)

        ref_ncluster = TORSIONSFINGERPRINTS_N_CLUSTER
        ref_populations = TORSIONSFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det["numClusters"]
        population = det["populations"].tolist()

        self.assertEqual(
            ncluster,
            ref_ncluster,
            msg="The number of cluster identified are not as expected",
        )
        self.assertEqual(
            population,
            ref_populations,
            msg="The population fo the cluster are not the expected one",
        )

    def test_cluster_circularFingerprints(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        _, det = cluster(mols, "circularFingerprints", returnDetails=True)

        ref_ncluster = CIRCULARFINGERPRINTS_N_CLUSTER
        ref_populations = CIRCULARFINGERPRINTS_POPULATION_CLUSTER

        ncluster = det["numClusters"]
        population = det["populations"].tolist()

        self.assertEqual(
            ncluster,
            ref_ncluster,
            msg="The number of cluster identified are not as expected",
        )
        self.assertEqual(
            population,
            ref_populations,
            msg="The population fo the cluster are not the expected one",
        )

    def test_cluster_shape(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        _, det = cluster(mols, "shape", returnDetails=True)

        n_clusters = det["numClusters"]

        self.assertIsInstance(n_clusters, np.int64, msg="None valid number of clusters")

    def test_cluster_mcs(self):
        sdffile = os.path.join(self.dataDir, "fda_drugs_light.sdf")
        lib = SmallMolLib(sdffile)
        mols = lib._mols

        _, det = cluster(mols, "mcs", returnDetails=True)

        ref_ncluster = MCS_N_CLUSTER
        ref_populations = MCS_POPULATION_CLUSTER

        ncluster = det["numClusters"]
        population = det["populations"].tolist()

        self.assertEqual(
            ncluster,
            ref_ncluster,
            msg="The number of cluster identified are not as expected",
        )
        self.assertEqual(
            population,
            ref_populations,
            msg="The population fo the cluster are not the expected one",
        )

    def test_restrained_embedding(self):
        from moleculekit.smallmol.tools.restrainedembed import restrainedEmbed
        import numpy as np

        atomPos = np.load(
            os.path.join(self.dataDir, "restrain_positions.npy"), allow_pickle=True
        ).item()
        refcoords = np.load(
            os.path.join(self.dataDir, "restrain_final_coords.npy"), allow_pickle=True
        )
        sml = SmallMolLib(os.path.join(self.dataDir, "restrain_molecule.sdf"))
        sm = sml[0]

        restrainedEmbed(sm._mol, atomPos)

        restr_atoms = np.sort(list(atomPos.keys()))
        coor_diff = refcoords[restr_atoms].squeeze() - sm._coords[restr_atoms].squeeze()
        assert np.abs(coor_diff).max() < 0.1

    # TODO: Add here tests for tautomers


if __name__ == "__main__":
    unittest.main(verbosity=2)



from moleculekit.molecule import Molecule
from pdb2pqr.aa import Amino
import numpy as np
import os


# PDB2PQR is quite finicky about the backbone coordinates so I copy them from ALA
backbone = Molecule().empty(6)
backbone.name[:] = ["N", "CA", "C", "O", "H", "HA"]
backbone.element[:] = ["N", "C", "C", "O", "H", "H"]
backbone.coords = (
    np.array(
        [
            [1.2010, 0.8470, 0.0000],
            [0.0000, 0.0000, 0.0000],
            [-1.250, 0.8810, 0.0000],
            [-2.185, 0.6600, -0.784],
            [1.2010, 1.8470, 0.0000],
            [0.0000, -0.557, -0.831],
        ]
    )
    .reshape((-1, 3, 1))
    .astype(np.float32)
    .copy()
)
backbone.bonds = np.array([[0, 1], [0, 4], [1, 5], [1, 2], [2, 3]])
backbone.bondtype = np.array(["1", "1", "1", "1", "2"], dtype=object)


def _get_residue_mol(mol: Molecule, nsres: str):
    from rdkit import Chem
    from aceprep.detector import getLigandFromDict
    from aceprep.prepare import RDKprepare
    import tempfile
    import logging

    acepreplog = logging.getLogger("aceprep")
    oldlevel = acepreplog.getEffectiveLevel()
    acepreplog.setLevel("CRITICAL")

    try:
        assert np.all(np.isin(["N", "CA", "C", "O"], mol.name))

        with tempfile.TemporaryDirectory() as outdir:
            resfile = os.path.join(outdir, f"residue_{nsres}.pdb")
            mol.write(resfile)

            outsdf = os.path.join(outdir, f"residue_{nsres}_templated.sdf")
            new_mol = getLigandFromDict(resfile, nsres)
            w = Chem.SDWriter(outsdf)
            w.write(new_mol)
            w.close()

            outsdfh = outsdf.replace(".sdf", "_h.sdf")
            RDKprepare(
                outsdf, outsdfh, os.path.join(outdir, "aceprep.log"), gen3d=False
            )

            mol = Molecule(outsdfh)
    except Exception:
        acepreplog.setLevel(oldlevel)
        raise

    acepreplog.setLevel(oldlevel)
    return mol


def _generate_custom_residue(inmol: Molecule, resname: str):
    from moleculekit.tools.graphalignment import makeMolGraph, compareGraphs

    import networkx as nx

    mol = _get_residue_mol(inmol, resname)
    # Adding an X_ to all atom names to separate them from the matched names
    for i, nn in enumerate(mol.name):
        mol.name[i] = f"X_{nn}"

    fields = ("element",)
    g1 = makeMolGraph(mol, "all", fields)
    g2 = makeMolGraph(inmol, "all", fields)
    _, _, matching = compareGraphs(
        g1, g2, fields=fields, tolerance=0.5, returnmatching=True
    )
    for pp in matching:  # Rename atoms in reference molecule
        mol.name[pp[0]] = inmol.name[pp[1]]

    # Cleaning up the reference molecule
    gg = nx.from_edgelist(mol.bonds)

    # Rename CA hydrogen to HA
    ca_idx = np.where(mol.name == "CA")[0][0]
    hidx = [i for i in gg.neighbors(ca_idx) if mol.element[i] == "H"]
    mol.name[hidx] = "HA"

    # Remove hydrogen attached to C
    c_idx = np.where(mol.name == "C")[0][0]
    hidx = [i for i in gg.neighbors(c_idx) if mol.element[i] == "H"]
    mol.remove(f"index {hidx[0]}", _logger=False)

    # Find all hydrogens attached to the backbone nitrogen
    n_idx = np.where(mol.name == "N")[0][0]
    hidx = [i for i in gg.neighbors(n_idx) if mol.element[i] == "H"]
    # Rename hydrogen bound to backbone N
    mol.name[hidx[-1]] = "H"
    # Rotate the last one to the correct rotamer
    mol.setDihedral(
        [np.where(mol.name == x)[0][0] for x in ["H", "N", "CA", "C"]], np.deg2rad(180)
    )
    # Keep only one H attached to backbone N atom
    mol.remove(f"index {' '.join(map(str, hidx[:2]))}", _logger=False)

    # Rename all added hydrogens
    hydr = mol.name == "X_H"
    mol.name[hydr] = [f"H{i}" for i in range(1, sum(hydr) + 1)]

    # Align and replace the backbone with the standard backbone of pdb2pqr
    mol.align("name N CA C", refmol=backbone)
    mol.remove("name N CA C O H HA", _logger=False)
    mol.insert(backbone, index=0)
    # Add back CA CB bond
    mol.bonds = np.vstack(([1, np.where(mol.name == "CB")[0][0]], mol.bonds))
    mol.bondtype = np.hstack(("1", mol.bondtype))
    # Rename residues
    mol.resname[:] = resname
    # AMBER format is: N H CA HA [sidechain] C O
    mol.reorderAtoms([0, 4, 1, 5] + list(range(6, mol.numAtoms)) + [2, 3])

    return mol


class CustomResidue(Amino):
    """Custom residue class. Hack for pdb2pqr which requires one class per residue in pdb2pqr.aa module"""

    def __init__(self, atoms, ref):
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letter_code(self):
        return "X"


def _mol_to_xml_def(mol: Molecule, outfile: str):
    with open(outfile, "w") as f:
        f.write("<?xml version='1.0'?>\n")
        f.write("<aminoacids>\n")
        f.write("  <residue>\n")
        f.write(f"    <name>{mol.resname[0]}</name>\n")

        for i in range(mol.numAtoms):
            f.write("    <atom>")
            f.write(
                f"""
      <name>{mol.name[i]}</name>
      <x>{mol.coords[i, 0, 0]:.3f}</x>
      <y>{mol.coords[i, 1, 0]:.3f}</y>
      <z>{mol.coords[i, 2, 0]:.3f}</z>\n"""
            )
            atombonds = mol.bonds[np.any(mol.bonds == i, axis=1), :].flatten()
            for b in sorted(atombonds):
                if b == i:
                    continue
                f.write(f"      <bond>{mol.name[b]}</bond>\n")
            f.write("    </atom>\n")
        f.write("  </residue>\n")
        f.write("</aminoacids>\n")


def _mol_to_dat_def(mol: Molecule, outfile: str):
    with open(outfile, "w") as f:
        for i in range(mol.numAtoms):
            radius = 0
            f.write(
                f"{mol.resname[i]}\t{mol.name[i]}\t{mol.charge[i]:.3f}\t{radius:.1f}\n"
            )


def _generate_mol2(outsdfh):
    from subprocess import call

    tmpmol = Molecule(outsdfh)
    nc = int(tmpmol.charge.sum())

    nshmol2 = outsdfh.replace(".sdf", ".mol2")
    call(
        [
            "antechamber",
            "-i",
            outsdfh,
            "-fi",
            "sdf",
            "-o",
            nshmol2,
            "-fo",
            "mol2",
            "-c",  # charge guessing method
            "gas",
            "-at",  # atom types
            "gaff2",
            "-s",  # verbosity
            "0",
            "-nc",  # net charge
            str(nc),
            "-j",  # atom type and bond type prediction index. 1: atom type (only)
            "1",
            "-pf",  # remove intermediate files
            "y",
        ]
    )
    frcmodfile = nshmol2.replace(".mol2", ".frcmod")
    call(
        [
            "parmchk2",
            "-i",
            nshmol2,
            "-o",
            frcmodfile,
            "-f",
            "mol2",
        ]
    )
    # with tempfile.TemporaryDirectory() as tmpdir:
    #     leapin = os.path.join(tmpdir, "tleap.in")
    #     leaptempl = os.path.join(
    #         home(), "newprep/build_templates/non-standard-aa/tleap.in"
    #     )
    #     outlib = os.path.join(outdir, f"{ns}.lib")
    #     with open(leaptempl) as fin, open(leapin, "w") as fout:
    #         template = Template(fin.read())
    #         fout.write(
    #             template.render(
    #                 mol2file=nshmol2,
    #                 frcmodfile=frcmodfile,
    #                 molname=ns,
    #                 outlib=outlib,
    #             )
    #         )
    #     call([teleap, *teleapimportflags, "-f", leapin])


def _get_custom_ff(user_ff=None, molkit_ff=True):
    import pdb2pqr
    import pdb2pqr.aa
    import tempfile
    from moleculekit.home import home
    import shutil
    from glob import glob

    original = os.path.join(os.path.dirname(pdb2pqr.__file__), "dat")
    molkitcustom = home(shareDir=os.path.join("pdb2pqr", "residues"))

    custom_xml = []
    if molkit_ff:
        custom_xml += glob(os.path.join(molkitcustom, "*.xml"))
    if user_ff is not None:
        custom_xml += glob(os.path.join(user_ff, "*.xml"))

    custom_dat = []
    if molkit_ff:
        custom_dat += glob(os.path.join(molkitcustom, "*.dat"))
    if user_ff is not None:
        custom_dat += glob(os.path.join(user_ff, "*.dat"))

    custom_resnames = np.unique(
        [os.path.splitext(os.path.basename(x))[0] for x in custom_xml]
    )

    with tempfile.TemporaryDirectory() as tmpdir:
        shutil.copytree(original, os.path.join(tmpdir, "originals"))

        with open(os.path.join(tmpdir, "originals", "AA.xml"), "r") as fin:
            lines = fin.readlines()

        # Merge moleculekit custom residues into AA.xml
        for cxml in custom_xml:
            with open(cxml, "r") as fc:
                custom_lines = fc.readlines()[2:-1]  # Skip header and parent tag
            lines = lines[:-1] + custom_lines + [lines[-1]]

        # Write out the new AA.xml
        out_aa = os.path.join(tmpdir, "AA.xml")
        with open(out_aa, "w") as fout:
            for line in lines:
                fout.write(line)

        with open(os.path.join(tmpdir, "originals", "PARSE.DAT"), "r") as fin:
            lines = fin.readlines()
            lines[-1] = lines[-1].strip() + "\n"

        # Merge moleculekit custom residues into PARSE.DAT
        for cdat in custom_dat:
            with open(cdat, "r") as fc:
                custom_lines = fc.readlines()
            lines += custom_lines

        # Write out the new PARSE.DAT
        out_dat = os.path.join(tmpdir, "PARSE.DAT")
        with open(out_dat, "w") as fout:
            for line in lines:
                fout.write(line)

        # HACK: pdb2pqr currently requires each residue to have a unique class in pdb2pqr.aa module
        for resn in custom_resnames:
            pdb2pqr.aa.__dict__[resn] = CustomResidue

        from pdb2pqr.io import get_definitions
        from pdb2pqr import forcefield

        definition = get_definitions(aa_path=out_aa)
        ff = forcefield.Forcefield("parse", definition, out_dat, None)
    return definition, ff



# (c) 2015-2018 Acellera Ltd http://www.acellera.com
# All Rights Reserved
# Distributed under HTMD Software License Agreement
# No redistribution in whole or part
#
import numpy as np
import logging
from moleculekit.molecule import Molecule
import unittest

logger = logging.getLogger(__name__)


def _get_sequence(mol: Molecule, sel):
    sel_mask = mol.atomselect(sel)
    protein_mask = mol.atomselect("protein")
    nucleic_mask = mol.atomselect("nucleic")
    if any(protein_mask & sel_mask) and any(nucleic_mask & sel_mask):
        raise RuntimeError(
            "Your selection contains both protein and nucleic residues. You need to clarify which selection to align."
        )
    molseg = "protein" if any(protein_mask) else "nucleic"
    seqmol, seqidx = mol.sequence(noseg=True, return_idx=True, sel=sel)
    seqidx = seqidx[molseg]
    seqmol = seqmol[molseg]
    segment_type = molseg
    return seqmol, seqidx, segment_type


def _get_max_align_region(molaln, refaln, nalignfragment):
    dsig = np.hstack(([False], (refaln != "-") & (molaln != "-"), [False])).astype(int)
    dsigdiff = np.diff(dsig)
    startIndex = np.where(dsigdiff > 0)[0]
    endIndex = np.where(dsigdiff < 0)[0]
    duration = endIndex - startIndex
    duration_sorted = np.sort(duration)[::-1]

    _list_starts = []
    _list_finish = []
    for n in range(nalignfragment):
        if n == len(duration):
            break
        idx = np.where(duration == duration_sorted[n])[0]
        start = startIndex[idx][0]
        finish = endIndex[idx][0]
        _list_starts.append(start)
        _list_finish.append(finish)
    return _list_starts, _list_finish


def sequenceStructureAlignment(
    mol,
    ref,
    molseg=None,
    refseg=None,
    molsel="all",
    refsel="all",
    maxalignments=10,
    nalignfragment=1,
):
    """Aligns two structures by their longests sequences alignment

    Parameters
    ----------
    mol : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The Molecule we want to align
    ref : :class:`Molecule <moleculekit.molecule.Molecule>` object
        The reference Molecule to which we want to align
    molseg : str
        The segment of `mol` we want to align
    refseg : str
        The segment of `ref` we want to align to
    maxalignments : int
        The maximum number of alignments we want to produce
    nalignfragment : int
        The number of fragments used for the alignment.

    Returns
    -------
    mols : list
        A list of Molecules each containing a different alignment.
    masks : list
        A list of boolean mask pairs of which atoms were aligned. [[molmask1, refmask1], [molmask2, refmask2], ...]
    """
    if molseg is not None:
        logger.warning(
            'molseg argument is deprecated. Please use molsel="segid X" instead.'
        )
        molsel = f"segid {molseg}"
    if refseg is not None:
        logger.warning(
            'refseg argument is deprecated. Please use refsel="segid X" instead.'
        )
        refsel = f"segid {refseg}"

    try:
        from Bio import pairwise2
    except ImportError:
        raise ImportError(
            "You need to install the biopython package to use this function. Try using `conda install biopython`."
        )
    from Bio.SubsMat import MatrixInfo as matlist

    if len([x for x in np.unique(mol.altloc) if len(x)]) > 1:
        raise RuntimeError(
            "Alternative atom locations detected in `mol`. Please remove these before calling this function."
        )
    if len([x for x in np.unique(ref.altloc) if len(x)]) > 1:
        raise RuntimeError(
            "Alternative atom locations detected in `ref`. Please remove these before calling this function."
        )

    seqmol, molidx, segment_type_mol = _get_sequence(mol, molsel)
    seqref, refidx, segment_type_ref = _get_sequence(ref, refsel)
    if segment_type_mol != segment_type_ref:
        raise RuntimeError(
            f"The segments of mol are of type {segment_type_mol} while for ref they are of type {segment_type_ref}. Please choose different segments to align."
        )
    segment_type = segment_type_mol

    if segment_type == "protein":
        alignments = pairwise2.align.globaldx(seqref, seqmol, matlist.blosum62)
    elif segment_type == "nucleic":
        alignments = pairwise2.align.globalxx(seqref, seqmol)
    numaln = len(alignments)

    if numaln > maxalignments:
        logger.warning(
            f"{numaln} alignments found. Limiting to {maxalignments} as specified in the `maxalignments` argument."
        )

    def _find_common_atom(mol, ref, idxmol, idxref):
        moln = mol.name[idxmol].tolist()
        refn = ref.name[idxref].tolist()
        prot_names = ["CA", "C", "O", "N"]
        nucl_names = ["C3'", "C4'", "C1'", "P"]
        names = prot_names if segment_type == "protein" else nucl_names
        for name in names:
            if name in moln and name in refn:
                return idxmol[moln.index(name)], idxref[refn.index(name)]
        raise RuntimeError(
            f"Could not find any of {names} in both mol and ref residues"
        )

    alignedstructs = []
    masks = []
    for i in range(min(maxalignments, numaln)):
        refaln = np.array(list(alignments[i][0]))
        molaln = np.array(list(alignments[i][1]))

        # By doing cumsum we calculate how many letters were before the current letter (i.e. residues before current)
        residref = np.cumsum(refaln != "-") - 1  # Start them from 0
        residmol = np.cumsum(molaln != "-") - 1  # Start them from 0

        # Find the region of maximum alignment between the molecules
        starts, ends = _get_max_align_region(molaln, refaln, nalignfragment)

        _refidx = []
        _molidx = []
        startends = []
        for ss, ff in zip(starts, ends):
            for k, (rr, mr) in enumerate(zip(residref[ss:ff], residmol[ss:ff])):
                mi, ri = _find_common_atom(mol, ref, molidx[mr], refidx[rr])
                _refidx.append(ri)
                _molidx.append(mi)
                # The rest is just for pretty printing
                if k == 0:
                    start_resid = mol.resid[mi]
                if k == ff - ss - 1:
                    end_resid = mol.resid[mi]
            startends.append(f"{start_resid}-{end_resid}")

        molmask = np.zeros(mol.numAtoms, dtype=bool)
        molmask[_molidx] = True
        refmask = np.zeros(ref.numAtoms, dtype=bool)
        refmask[_refidx] = True

        logger.info(
            f"Alignment #{i} was done on {len(_refidx)} residues: {' ,'.join(startends)}"
        )

        alignedmol = mol.copy()
        alignedmol.align(sel=molmask, refmol=ref, refsel=refmask)
        alignedstructs.append(alignedmol)
        masks.append([molmask, refmask])

    return alignedstructs, masks


class _TestSequenceStructuralAlignment(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        from moleculekit.home import home

        self.home = home(dataDir="test-sequence-alignment")

    def test_align_protein(self):
        import os

        mol1 = Molecule(os.path.join(self.home, "4OBE.pdb"))
        mol2 = Molecule(os.path.join(self.home, "6OB2.pdb"))
        refmol = Molecule(os.path.join(self.home, "4OBE_aligned.pdb"))
        mol3, _ = sequenceStructureAlignment(
            mol1, mol2, molsel="protein", refsel="protein"
        )

        assert np.allclose(refmol.coords, mol3[0].coords, atol=1e-3)

    def test_align_rna(self):
        import os

        mol1 = Molecule(os.path.join(self.home, "5C45.pdb"))
        mol2 = Molecule(os.path.join(self.home, "5C45_sim.pdb"))
        refmol = Molecule(os.path.join(self.home, "5C45_aligned.pdb"))
        mol3, _ = sequenceStructureAlignment(mol1, mol2)

        assert np.allclose(refmol.coords, mol3[0].coords, atol=1e-3)


if __name__ == "__main__":
    unittest.main(verbosity=2)
